# -*- coding: utf-8 -*-

import inspect

from brainpy import errors
from brainpy.math import profile

__all__ = [
  'get_args',
  'check_kws',
]


def check_kws(parameters, keywords):
  for key, meaning in keywords.items():
    if key in parameters:
      raise errors.CodeError(f'"{key}" is a keyword for '
                             f'numerical solvers in BrainPy, denoting '
                             f'{meaning}. Please change another name.')


def get_args(f):
  """Get the function arguments.

  >>> def f1(a, b, t, *args, c=1): pass
  >>> get_args(f1)
  (['a', 'b'], ['t', '*args', 'c'], ['a', 'b', 't', '*args', 'c=1'])

  >>> def f2(a, b, *args, c=1, **kwargs): pass
  >>> get_args(f2)
  ValueError: Don not support dict of keyword arguments: **kwargs

  >>> def f3(a, b, t, c=1, d=2): pass
  >>> get_args(f4)
  (['a', 'b'], ['t', 'c', 'd'], ['a', 'b', 't', 'c=1', 'd=2'])

  >>> def f4(a, b, t, *args): pass
  >>> get_args(f4)
  (['a', 'b'], ['t', '*args'], ['a', 'b', 't', '*args'])

  >>> scope = {}
  >>> exec(compile('def f5(a, b, t, *args): pass', '', 'exec'), scope)
  >>> get_args(scope['f5'])
  (['a', 'b'], ['t', '*args'], ['a', 'b', 't', '*args'])

  Parameters
  ----------
  f : callable
      The function.

  Returns
  -------
  args : tuple
      The variable names, the other arguments, and the original args.
  """

  # 1. get the function arguments
  reduced_args = []
  original_args = []

  for name, par in inspect.signature(f).parameters.items():
    if par.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
      reduced_args.append(par.name)

    elif par.kind is inspect.Parameter.VAR_POSITIONAL:
      reduced_args.append(f'*{par.name}')

    elif par.kind is inspect.Parameter.KEYWORD_ONLY:
      reduced_args.append(par.name)

    elif par.kind is inspect.Parameter.POSITIONAL_ONLY:
      raise errors.DiffEqError('Don not support positional only parameters, e.g., /')
    elif par.kind is inspect.Parameter.VAR_KEYWORD:
      raise errors.DiffEqError(f'Don not support dict of keyword arguments: {str(par)}')
    else:
      raise errors.DiffEqError(f'Unknown argument type: {par.kind}')

    original_args.append(str(par))

  # 2. analyze the function arguments
  #   2.1 class keywords
  class_kw = []
  if reduced_args[0] in profile.CLASS_KEYWORDS:
    class_kw.append(reduced_args[0])
    reduced_args = reduced_args[1:]
  for a in reduced_args:
    if a in profile.CLASS_KEYWORDS:
      raise errors.DiffEqError(f'Class keywords "{a}" must be defined '
                               f'as the first argument.')
  #  2.2 variable names
  var_names = []
  for a in reduced_args:
    if a == 't':
      break
    var_names.append(a)
  else:
    raise ValueError('Do not find time variable "t".')
  other_args = reduced_args[len(var_names):]
  return class_kw, var_names, other_args, original_args



# -*- coding: utf-8 -*-

from pprint import pprint

from brainpy import errors
from brainpy.integrators import constants

_ODE_UNKNOWN_NO = 0


def f_names(f):
  func_name = constants.unique_name('ode')
  if f.__name__.isidentifier():
    func_name += '_' + f.__name__
  return func_name


def step(class_kw, vars, dt_var, A, C, code_lines, other_args):
  # steps
  for si, sval in enumerate(A):
    # k-step arguments
    k_args = []
    for v in vars:
      k_arg = f'{v}'
      for j, sv in enumerate(sval):
        if sv not in [0., '0.0', '0.', '0']:
          if sv in ['1.0', '1.', '1', 1.]:
            k_arg += f' + {dt_var} * d{v}_k{j + 1}'
          else:
            k_arg += f' + {dt_var} * d{v}_k{j + 1} * {sv}'
      if k_arg != v:
        name = f'k{si + 1}_{v}_arg'
        code_lines.append(f'  {name} = {k_arg}')
        k_args.append(name)
      else:
        k_args.append(v)

    t_arg = 't'
    if C[si] not in [0., '0.', '0']:
      if C[si] in ['1.', '1', 1.]:
        t_arg += f' + {dt_var}'
      else:
        t_arg += f' + {dt_var} * {C[si]}'
      name = f'k{si + 1}_t_arg'
      code_lines.append(f'  {name} = {t_arg}')
      k_args.append(name)
    else:
      k_args.append(t_arg)

    # k-step derivative names
    k_derivatives = [f'd{v}_k{si + 1}' for v in vars]

    # k-step code line
    code_lines.append(f'  {", ".join(k_derivatives)} = f('
                      f'{", ".join(class_kw + k_args + other_args[1:])})')


def update(vars, dt_var, B, code_lines):
  return_args = []
  for v in vars:
    result = v
    for i, b1 in enumerate(B):
      if b1 not in [0., '0.', '0']:
        result += f' + d{v}_k{i + 1} * {dt_var} * {b1}'
    code_lines.append(f'  {v}_new = {result}')
    return_args.append(f'{v}_new')
  return return_args


def compile_and_assign_attrs(raw_func,
                             code_lines,
                             code_scope,
                             func_name,
                             variables,
                             parameters,
                             dt,
                             var_type,
                             method,
                             show_code=False):
  if not isinstance(raw_func, dict):
    raise errors.IntegratorError('"raw_func" must be a dict of functions.')

  code_scope_old = {key: val for key, val in code_scope.items()}

  # compile functions
  code = '\n'.join(code_lines)
  if show_code:
    print(code)
    print()
    pprint(code_scope)
    print()
  exec(compile(code, '', 'exec'), code_scope)
  new_f = code_scope[func_name]

  # assign values
  new_f.brainpy_data = dict(raw_func=raw_func,
                            code_lines=code_lines,
                            code_scope=code_scope_old,
                            variables=variables,
                            parameters=parameters,
                            method=method,
                            dt=dt,
                            func_name=func_name,
                            var_type=var_type)

  return new_f



# -*- coding: utf-8 -*-

import ast
from pprint import pprint

import pytest

from brainpy.errors import DiffEqError
from brainpy.integrators.analysis_by_ast import DiffEqReader
from brainpy.integrators.analysis_by_ast import separate_variables


def test_reader1():
    eq_code = '''
def func():
    a, b = f()
    c = a + b
    d = a * b
    return c, d
    '''
    analyser = DiffEqReader()
    analyser.visit(ast.parse(eq_code))

    print(analyser.returns)
    assert analyser.returns == ['c', 'd']

    print(analyser.code_lines)
    assert analyser.code_lines == ['a, b = f()\n',
                                   'c = a + b\n',
                                   'd = a * b\n']

    print(analyser.rights)
    assert analyser.rights == ['f()', 'a + b', 'a * b']

    print(analyser.variables)
    assert analyser.variables == [['a', 'b'],
                                  ['c'],
                                  ['d']]


def test_reader2():
    eq_code = '''
def func():
    a, b = f()
    if a > 0:
        c = a + b
    else:
        c = a * b
    return c
        '''
    analyser = DiffEqReader()
    with pytest.raises(DiffEqError):
        analyser.visit(ast.parse(eq_code))


def test_reader3():
    eq_code = '''
def func():
    a, b = f()
    for i in range(10):
        a += b
    return a
        '''
    analyser = DiffEqReader()
    with pytest.raises(DiffEqError):
        analyser.visit(ast.parse(eq_code))


def test_separate_variables1():
    eq_code = '''
def integral(V, m, h, n, t, Iext):
    alpha = 0.1 * (V + 40) / (1 - np.exp(-(V + 40) / 10))
    beta = 4.0 * np.exp(-(V + 65) / 18)
    dmdt = alpha * (1 - m) - beta * m

    alpha = 0.07 * np.exp(-(V + 65) / 20.)
    beta = 1 / (1 + np.exp(-(V + 35) / 10))
    dhdt = alpha * (1 - h) - beta * h
    return dmdt, dhdt
'''
    analyser = DiffEqReader()
    analyser.visit(ast.parse(eq_code))

    print('returns: ')
    pprint(analyser.returns)

    print("code_lines: ")
    pprint(analyser.code_lines)

    print("rights: ")
    pprint(analyser.rights)

    print("variables: ")
    pprint(analyser.variables)

    r = separate_variables(eq_code)
    pprint(r)


def test_separate_variables2():
    code = '''def derivative(V, m, h, n, t, C, gNa, ENa, gK, EK, gL, EL, Iext):
    alpha = 0.1 * (V + 40) / (1 - bp.backend.exp(-(V + 40) / 10))
    beta = 4.0 * bp.backend.exp(-(V + 65) / 18)
    dmdt = alpha * (1 - m) - beta * m

    alpha = 0.07 * bp.backend.exp(-(V + 65) / 20.)
    beta = 1 / (1 + bp.backend.exp(-(V + 35) / 10))
    dhdt = alpha * (1 - h) - beta * h

    alpha = 0.01 * (V + 55) / (1 - bp.backend.exp(-(V + 55) / 10))
    beta = 0.125 * bp.backend.exp(-(V + 65) / 80)
    dndt = alpha * (1 - n) - beta * n

    I_Na = (gNa * m ** 3.0 * h) * (V - ENa)
    I_K = (gK * n ** 4.0) * (V - EK)
    I_leak = gL * (V - EL)
    dVdt = (- I_Na - I_K - I_leak + Iext) / C

    return dVdt, dmdt, dhdt, dndt
    '''

    from pprint import pprint
    pprint(separate_variables(code))

# test_reader1()
# test_reader2()
# test_reader3()
# test_separate_variables1()
# test_separate_variables2()



# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt

import brainpy as bp


def test_GaussianDecay1():
  bp.math.use_backend('numpy')
  conn = bp.initialize.GaussianDecay(sigma=4, max_w=1., periodic_boundary=True)
  weights = conn(100)

  # plt.imshow(weights)
  # plt.show()


def test_GaussianDecay2():
  bp.math.use_backend('numpy')

  size = (30, 30)
  conn = bp.initialize.GaussianDecay(sigma=4, max_w=1., periodic_boundary=True)
  weights = conn(size)

  # plt.imshow(weights)
  # plt.show()
  #
  # plt.imshow(weights[0].reshape(size))
  # plt.show()


def test_GaussianDecay3():
  bp.math.use_backend('numpy')

  size = (20, 20, 20)
  conn = bp.initialize.GaussianDecay(sigma=4, max_w=1., periodic_boundary=True)
  weights = conn(size)

  # plt.imshow(weights)
  # plt.show()


def test_DOGDecay1():
  bp.math.use_backend('numpy')
  conn = bp.initialize.DOGDecay(sigmas=(1., 2.5), max_ws=(1.0, 0.7), periodic_boundary=True)
  weights = conn(100)

  # pos = plt.imshow(weights, cmap='cool', interpolation='none')
  # plt.colorbar(pos)
  # plt.show()


def test_DOGDecay2():
  bp.math.use_backend('numpy')

  size = (30, 30)
  conn = bp.initialize.DOGDecay(sigmas=(1., 2.5), max_ws=(1.0, 0.7), periodic_boundary=True)
  weights = conn(size)

  # pos = plt.imshow(weights, cmap='cool', interpolation='none')
  # plt.colorbar(pos)
  # plt.show()

  # pos = plt.imshow(weights[0].reshape(size), cmap='cool', interpolation='none')
  # plt.colorbar(pos)
  # plt.show()


def test_DOGDecay3():
  bp.math.use_backend('numpy')

  size = (10, 10, 10)
  conn = bp.initialize.DOGDecay(sigmas=(1., 2.5), max_ws=(1.0, 0.7), periodic_boundary=True)
  weights = conn(size)

  # pos = plt.imshow(weights, cmap='cool', interpolation='none')
  # plt.colorbar(pos)
  # plt.show()



# -*- coding: utf-8 -*-

# necessity to wrap the jax.numpy.ndarray:
# 1. for parameters and variables which want to
#    modify in the JIT mode, this wrapper is necessary.
#    Because, directly change the value in the "self.xx"
#    cannot work.
# 2. In JAX, ndarray is immutable. This wrapper make
#    the index update is the same way with the numpy
#


from . import activations
from . import fft
from . import linalg
from . import losses
from . import optimizers
from . import random

from .activations import *
from .code import *
from .controls import *
from .compilation import *
from .function import *
from .gradient import *
from .jaxarray import *
from .ops import *



# -*- coding: utf-8 -*-


import brainpy as bp

bp.math.use_backend('numpy')
bp.integrators.set_default_odeint(method='rk4')


class LIF(bp.NeuGroup):
  def __init__(self, size, t_refractory=1., V_rest=0., V_reset=-5.,
               V_th=20., R=1., tau=10., **kwargs):
    super(LIF, self).__init__(size=size, **kwargs)

    # parameters
    self.V_rest = V_rest
    self.V_reset = V_reset
    self.V_th = V_th
    self.R = R
    self.tau = tau
    self.t_refractory = t_refractory

    # variables
    self.V = bp.math.Variable(bp.math.ones(self.num) * V_rest)
    self.input = bp.math.Variable(bp.math.zeros(self.num))
    self.refractory = bp.math.Variable(bp.math.zeros(self.num, dtype=bool))
    self.spike = bp.math.Variable(bp.math.zeros(self.num, dtype=bool))
    self.t_last_spike = bp.math.Variable(bp.math.ones(self.num) * -1e7)

  @bp.odeint
  def integral(self, V, t, Iext):
    dvdt = (-V + self.V_rest + self.R * Iext) / self.tau
    return dvdt

  def update(self, _t, _dt):
    refractory = (_t - self.t_last_spike) <= self.t_refractory
    V = self.integral(self.V, _t, self.input, dt=_dt)
    V = bp.math.where(refractory, self.V, V)
    spike = self.V_th <= V
    self.t_last_spike[:] = bp.math.where(spike, _t, self.t_last_spike)
    self.V[:] = bp.math.where(spike, self.V_reset, V)
    self.refractory[:] = bp.math.logical_or(refractory, spike)
    self.input[:] = 0.
    self.spike[:] = spike


class HH(bp.NeuGroup):
  def __init__(self, size, ENa=50., gNa=120., EK=-77., gK=36., EL=-54.387,
               gL=0.03, V_th=20., C=1.0, **kwargs):
    super(HH, self).__init__(size=size, **kwargs)

    # parameters
    self.ENa = ENa
    self.EK = EK
    self.EL = EL
    self.gNa = gNa
    self.gK = gK
    self.gL = gL
    self.C = C
    self.V_th = V_th

    # variables
    self.V = bp.math.Variable(-65. * bp.math.ones(self.num))
    self.m = bp.math.Variable(0.5 * bp.math.ones(self.num))
    self.h = bp.math.Variable(0.6 * bp.math.ones(self.num))
    self.n = bp.math.Variable(0.32 * bp.math.ones(self.num))
    self.input = bp.math.Variable(bp.math.zeros(self.num))
    self.spike = bp.math.Variable(bp.math.zeros(self.num, dtype=bool))
    self.t_last_spike = bp.math.Variable(bp.math.ones(self.num) * -1e7)

  @bp.odeint
  def integral(self, V, m, h, n, t, Iext):
    alpha = 0.1 * (V + 40) / (1 - bp.math.exp(-(V + 40) / 10))
    beta = 4.0 * bp.math.exp(-(V + 65) / 18)
    dmdt = alpha * (1 - m) - beta * m

    alpha = 0.07 * bp.math.exp(-(V + 65) / 20.)
    beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))
    dhdt = alpha * (1 - h) - beta * h

    alpha = 0.01 * (V + 55) / (1 - bp.math.exp(-(V + 55) / 10))
    beta = 0.125 * bp.math.exp(-(V + 65) / 80)
    dndt = alpha * (1 - n) - beta * n

    I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
    I_K = (self.gK * n ** 4.0) * (V - self.EK)
    I_leak = self.gL * (V - self.EL)
    dVdt = (- I_Na - I_K - I_leak + Iext) / self.C

    return dVdt, dmdt, dhdt, dndt

  def update(self, _t, _dt):
    V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input, dt=_dt)
    self.spike[:] = bp.math.logical_and(self.V < self.V_th, V >= self.V_th)
    self.V[:] = V
    self.m[:] = m
    self.h[:] = h
    self.n[:] = n
    self.input[:] = 0.


def test_jit_neugroup_lif():
  bp.math.jit(LIF(10), show_code=True).run(1.)


def test_jit_neugroup_lif_bounded_func():
  bp.integrators.set_default_odeint(method='exponential_euler')
  bp.math.jit(LIF(10).update, show_code=True)


def test_jit_neugroup_lif_integrator():
  bp.integrators.set_default_odeint(method='exponential_euler')
  bp.math.jit(LIF(10).integral, show_code=True)


def test_jit_neugroup_hh():
  bp.math.jit(HH(10), show_code=True).run(1.)


def test_jit_neugroup_hh_bounded_func():
  bp.integrators.set_default_odeint(method='exponential_euler')
  bp.math.jit(HH(10).update, show_code=True)


def test_jit_neugroup_hh_integrator():
  bp.integrators.set_default_odeint(method='exponential_euler')
  bp.math.jit(HH(10).integral, show_code=True)


class LogisticRegression(bp.Base):
  def __init__(self, dimension):
    super(LogisticRegression, self).__init__()

    self.dimension = dimension
    self.w = bp.math.Variable(2.0 * bp.math.ones(dimension) - 1.3)

  def __call__(self, X, Y):
    u = bp.math.dot(((1.0 / (1.0 + bp.math.exp(-Y * bp.math.dot(X, self.w))) - 1.0) * Y), X)
    self.w[:] -= u


def test1():
  num_dim = 10
  num_points = 200
  # num_points = 20000000
  points = bp.math.random.random((num_points, num_dim))
  labels = bp.math.random.random(num_points)
  lr = LogisticRegression(num_dim)
  jit_lr = bp.math.jit(lr, show_code=True)
  print(lr.w)

  num_iter = 20
  for i in range(num_iter):
    jit_lr(points, labels)
  print(lr.w)





# -*- coding: utf-8 -*-


class RunningError(Exception):
  """The error occurred in the running function."""
  pass


class BrainPyError(Exception):
  """General BrainPy error."""
  pass


class IntegratorError(Exception):
  pass


class DiffEqError(Exception):
  """The differential equation definition error.
  """
  pass


class CodeError(Exception):
  """Code definition error.
  """
  pass


class AnalyzerError(Exception):
  """Differential equation analyzer error.
  """


class PackageMissingError(Exception):
  """The package missing error.
  """
  pass


class BackendNotInstalled(Exception):
  def __init__(self, backend):
    super(BackendNotInstalled, self).__init__(
      '"{bk}" must be installed when the user wants to use {bk} backend. \n'
      'Please install {bk} through "pip install {bk}" '
      'or "conda install {bk}".'.format(bk=backend))


class UniqueNameError(Exception):
  def __init__(self, *args):
    super(UniqueNameError, self).__init__(*args)


class UnsupportedError(Exception):
  pass


class NoLongerSupportError(Exception):
  pass




# -*- coding: utf-8 -*-

import numpy

try:
  import numba
except ModuleNotFoundError:
  numba = None

__all__ = [
  'numba_jit',
  'numba_seed',
]


def numba_jit(f=None, **setting):
  if numba is None:
    if f is None: return lambda f: f
    else: return f
  else:
    if f is None: return lambda f: numba.njit(f, **setting)
    else: return numba.njit(f, **setting)


@numba_jit
def _numba_seed(seed=None):
  numpy.random.seed(seed)


def numba_seed(seed):
  if numba is None:
    pass
  else:
    if seed is None: seed = numpy.random.randint(0, 1000000)
    assert isinstance(seed, int)
    _numba_seed(seed)




# -*- coding: utf-8 -*-

import pytest
import matplotlib.pyplot as plt
import numba
import numpy as np
from mpl_toolkits.mplot3d import Axes3D


import brainpy as bp
from brainpy.integrators import sde


sigma = 10
beta = 8 / 3
rho = 28
p = 0.1


@numba.njit
def lorenz_f(x, y, z, t):
    dx = sigma * (y - x)
    dy = x * (rho - z) - y
    dz = x * y - beta * z
    return dx, dy, dz


@numba.njit
def lorenz_g(x, y, z, t):
    return p * x, p * y, p * z


def lorenz_system(method, **kwargs):
    integral = numba.njit(method(f=lorenz_f,
                                 g=lorenz_g,
                                 show_code=True,
                                 dt=0.005,
                                 **kwargs))

    times = np.arange(0, 100, 0.01)
    mon1 = []
    mon2 = []
    mon3 = []
    x, y, z = 1, 1, 1
    for t in times:
        x, y, z = integral(x, y, z, t)
        mon1.append(x)
        mon2.append(y)
        mon3.append(z)
    mon1 = np.array(mon1)
    mon2 = np.array(mon2)
    mon3 = np.array(mon3)

    # fig = plt.figure()
    # ax = fig.gca(projection='3d')
    # plt.plot(mon1, mon2, mon3)
    # ax.set_xlabel('x')
    # ax.set_xlabel('y')
    # ax.set_xlabel('z')
    # plt.show()

def test():
    lorenz_system(sde.srk1w1_scalar, )
    with pytest.raises(bp.errors.IntegratorError):
        lorenz_system(sde.srk1w1_scalar, wiener_type=bp.integrators.VECTOR_WIENER)
    lorenz_system(sde.srk2w1_scalar)
    lorenz_system(sde.euler, intg_type=bp.integrators.ITO_SDE)
    lorenz_system(sde.euler, intg_type=bp.integrators.STRA_SDE)
    lorenz_system(sde.milstein, intg_type=bp.integrators.ITO_SDE)
    lorenz_system(sde.milstein, intg_type=bp.integrators.STRA_SDE)





# -*- coding: utf-8 -*-

import _thread as thread
import ast
import inspect
import re
import threading

import numpy as np

from brainpy import errors, math, tools
from brainpy.integrators import analysis_by_ast, odeint, utils, DE_INT

try:
  import numba
  from numba.core.dispatcher import Dispatcher
except ModuleNotFoundError:
  numba = None
  Dispatcher = None

DynamicalSystem = None

__all__ = [
  'transform_integrals_to_model',
  'DynamicModel',
  'rescale',
  'timeout',
  'jit_compile',
  'add_arrow',
  'contain_unknown_symbol',
]

# Get functions in backend
_functions_in_math = []
for key in dir(math):
  if not key.startswith('__'):
    _functions_in_math.append(getattr(math, key))

# Get functions in NumPy
_functions_in_numpy = []
for key in dir(np):
  if not key.startswith('__'):
    _functions_in_numpy.append(getattr(np, key))
for key in dir(np.random):
  if not key.startswith('__'):
    _functions_in_numpy.append(getattr(np.random, key))
for key in dir(np.linalg):
  if not key.startswith('__'):
    _functions_in_numpy.append(getattr(np.linalg, key))


def func_in_numpy_or_math(func):
  return func in _functions_in_math or func in _functions_in_numpy


def transform_integrals(integrals, method='euler'):
  global DynamicalSystem
  if DynamicalSystem is None: from brainpy.simulation.brainobjects.base import DynamicalSystem

  pars_update = {}
  new_integrals = []
  for integral in integrals:
    # integral function
    if Dispatcher is not None and isinstance(integral, Dispatcher):
      integral = integral.py_func
    else:
      integral = integral

    # derivative function
    func = integral.brainpy_data['raw_func']['f']
    if Dispatcher is not None and isinstance(func, Dispatcher):
      func = func.py_func
    func_name = func.__name__

    # arguments
    class_kw, variables, parameters, _ = utils.get_args(func)

    if len(class_kw) == 0:
      pars_update.update({p: None for p in parameters[1:]})
      new_integrals.append(integral)

    else:
      assert len(class_kw) == 1
      code = tools.deindent(inspect.getsource(func)).strip()
      tree = ast.parse(code)
      tree.body[0].args.args.pop(0)  # remove "self" arg
      self_data = re.findall('\\b' + class_kw[0] + '\\.[A-Za-z_][A-Za-z0-9_.]*\\b', code)
      self_data = list(set(self_data))

      # node of integral
      f_node = None
      if hasattr(integral, '__self__') and isinstance(integral.__self__, DynamicalSystem):
        f_node = integral.__self__

      # node of derivative function
      func_node = None
      if f_node:
        func_node = f_node
      elif hasattr(func, '__self__') and isinstance(func.__self__, DynamicalSystem):
        func_node = func.__self__

      # code scope
      closure_vars = inspect.getclosurevars(func)
      code_scope = dict(closure_vars.nonlocals)
      code_scope.update(closure_vars.globals)

      # analyze variables and functions accessed by the self.xx
      if func_node:
        arguments = set()
        data_to_replace = {}
        for key in self_data:
          split_keys = key.split('.')
          if len(split_keys) < 2:
            raise errors.BrainPyError

          # get target and data
          target = func_node
          for i in range(1, len(split_keys)):
            next_target = getattr(target, split_keys[i])
            if not isinstance(next_target, DynamicalSystem):
              break
            target = next_target
          else:
            raise errors.BrainPyError
          data = getattr(target, split_keys[i])

          key = '.'.join(split_keys[:i + 1])
          if isinstance(data, np.random.RandomState):
            data_to_replace[key] = f'{target.name}_{split_keys[i]}'  # replace the data
            code_scope[f'{target.name}_{split_keys[i]}'] = np.random  # replace RandomState

          elif callable(data):
            data_to_replace[key] = f'{target.name}_{split_keys[i]}'  # replace the data
            code_scope[f'{target.name}_{split_keys[i]}'] = data

          elif isinstance(data, (math.ndarray, int, float)):
            assert len(split_keys) == i + 1
            arguments.add(split_keys[i])
            pars_update[split_keys[i]] = data
            data_to_replace[key] = split_keys[i]  # replace the data

          else:  # parameters
            data_to_replace[key] = f'{target.name}_{split_keys[i]}'  # replace the data
            code_scope[f'{target.name}_{split_keys[i]}'] = data

        # final code
        tree.body[0].decorator_list.clear()
        tree.body[0].args.args.extend([ast.Name(id=a) for a in sorted(arguments)])
        tree.body[0].args.defaults.extend([ast.Constant(None) for _ in sorted(arguments)])
        code = tools.ast2code(tree)
        code = tools.word_replace(code, data_to_replace, exclude_dot=False)

        # compile new function
        # if show_code:
        #   print(code)
        #   print()
        #   pprint(code_scope)
        #   print()
        exec(compile(code, '', 'exec'), code_scope)
        func = code_scope[func_name]
        func.code = code

        new_integrals.append(odeint(f=func, method=method))
  return new_integrals, pars_update


def transform_integrals_to_model(integrals, method='euler'):
  from brainpy.integrators import analysis_by_sympy
  global DynamicalSystem
  if DynamicalSystem is None: from brainpy.simulation.brainobjects.base import DynamicalSystem

  # check integrals
  if callable(integrals):
    integrals = [integrals]
  if isinstance(integrals, (list, tuple)):
    integrals = tuple(integrals)
  elif isinstance(integrals, dict):
    integrals = tuple(integrals.values())
  elif isinstance(integrals, DynamicalSystem):
    integrals = tuple(integrals.ints().unique().values())
  else:
    raise errors.UnsupportedError(f'Dynamics analysis by symbolic approach only supports '
                                  f'integrators, but we got: {type(integrals)}: {str(integrals)}')
  for intg in integrals:
    assert callable(intg) and intg.__name__.startswith(DE_INT)

  integrals, pars_update = transform_integrals(integrals, method=method)

  all_scope = dict(math=math)
  all_variables = set()
  all_parameters = set()
  analyzers = []
  for integral in integrals:
    # integral function
    if Dispatcher is not None and isinstance(integral, Dispatcher):
      integral = integral.py_func
    else:
      integral = integral

    # original function
    f = integral.brainpy_data['raw_func']['f']
    if Dispatcher is not None and isinstance(f, Dispatcher):
      f = f.py_func
    func_name = f.__name__

    # code scope
    closure_vars = inspect.getclosurevars(f)
    code_scope = dict(closure_vars.nonlocals)
    code_scope.update(dict(closure_vars.globals))

    # separate variables
    analysis = analysis_by_ast.separate_variables(f.code if hasattr(f, 'code') else f)
    variables_for_returns = analysis['variables_for_returns']
    expressions_for_returns = analysis['expressions_for_returns']
    for vi, (key, vars) in enumerate(variables_for_returns.items()):
      variables = []
      for v in vars:
        if len(v) > 1:
          raise ValueError('Cannot analyze multi-assignment code line.')
        variables.append(v[0])
      expressions = expressions_for_returns[key]
      var_name = integral.brainpy_data['variables'][vi]
      DE = analysis_by_sympy.SingleDiffEq(var_name=var_name,
                                          variables=variables,
                                          expressions=expressions,
                                          derivative_expr=key,
                                          scope=code_scope,
                                          func_name=func_name)
      analyzers.append(DE)

    # others
    for var in integral.brainpy_data['variables']:
      if var in all_variables:
        raise errors.BrainPyError(f'Variable {var} has been defined before. Cannot group '
                                  f'this integral as a dynamic system.')
      all_variables.add(var)
    all_parameters.update(integral.brainpy_data['parameters'])
    all_scope.update(code_scope)

  # form a dynamic model
  return DynamicModel(integrals=integrals,
                      analyzers=analyzers,
                      variables=list(all_variables),
                      parameters=list(all_parameters),
                      pars_update=pars_update,
                      scopes=all_scope)


class DynamicModel(object):
  def __init__(self, integrals, analyzers, variables, parameters, scopes,
               pars_update=None):
    self.integrals = integrals
    self.analyzers = analyzers
    self.variables = variables
    self.parameters = parameters
    self.scopes = scopes
    self.pars_update = pars_update


def rescale(min_max, scale=0.01):
  """Rescale lim."""
  min_, max_ = min_max
  length = max_ - min_
  min_ -= scale * length
  max_ += scale * length
  return min_, max_


def timeout(s):
  """Add a timeout parameter to a function and return it.

  Parameters
  ----------
  s : float
      Time limit in seconds.

  Returns
  -------
  func : callable
      Functional results. Or, raise an error of KeyboardInterrupt.
  """

  def outer(fn):
    def inner(*args, **kwargs):
      timer = threading.Timer(s, thread.interrupt_main)
      timer.start()
      try:
        result = fn(*args, **kwargs)
      finally:
        timer.cancel()
      return result

    return inner

  return outer


def _jit(func):
  if func_in_numpy_or_math(func):
    return func
  if isinstance(func, Dispatcher):
    return func
  vars = inspect.getclosurevars(func)
  code_scope = dict(vars.nonlocals)
  code_scope.update(vars.globals)

  modified = False
  # check scope variables
  for k, v in code_scope.items():
    # function
    if callable(v):
      if (not func_in_numpy_or_math(v)) and (not isinstance(v, Dispatcher)):
        code_scope[k] = _jit(v)
        modified = True

  if modified:
    func_code = tools.deindent(tools.get_func_source(func))
    exec(compile(func_code, '', "exec"), code_scope)
    func = code_scope[func.__name__]
    return numba.njit(func)
  else:
    return numba.njit(func)


def _is_numpy_bk():
  bk_name = math.get_backend_name()
  return bk_name.startswith('numba') or bk_name == 'numpy'


def jit_compile(scope, func_code, func_name):
  if (numba is None) or (not _is_numpy_bk()):
    func_scope = scope
  else:
    assert Dispatcher is not None
    # get function scope
    func_scope = dict()
    for key, val in scope.items():
      if callable(val):
        if func_in_numpy_or_math(val):
          pass
        elif isinstance(val, Dispatcher):
          pass
        else:
          val = _jit(val)
      func_scope[key] = val

  # compile function
  exec(compile(func_code, '', 'exec'), func_scope)
  if numba is None:
    return func_scope[func_name]
  else:
    return numba.njit(func_scope[func_name])


def contain_unknown_symbol(expr, scope):
  """Examine where the given expression ``expr`` has the unknown symbol in ``scope``.

  Returns
  -------
  res : bool
      True or False.
  """
  ids = tools.get_identifiers(expr)
  for id_ in ids:
    if '.' in id_:
      prefix = id_.split('.')[0].strip()
      if prefix not in scope:
        return True
    if id_ not in scope:
      return True
  return False


def add_arrow(line, position=None, direction='right', size=15, color=None):
  """
  add an arrow to a line.

  line:       Line2D object
  position:   x-position of the arrow. If None, mean of xdata is taken
  direction:  'left' or 'right'
  size:       size of the arrow in fontsize points
  color:      if None, line color is taken.
  """
  if color is None:
    color = line.get_color()

  xdata = line.get_xdata()
  ydata = line.get_ydata()

  if position is None:
    position = xdata.mean()
  # find closest index
  start_ind = np.argmin(np.absolute(xdata - position))
  if direction == 'right':
    end_ind = start_ind + 1
  else:
    end_ind = start_ind - 1

  line.axes.annotate(text='',
                     xytext=(xdata[start_ind], ydata[start_ind]),
                     xy=(xdata[end_ind], ydata[end_ind]),
                     arrowprops=dict(arrowstyle="->", color=color),
                     size=size)


@tools.numba_jit
def f1(arr, grad, tol):
  condition = np.logical_and(grad[:-1] * grad[1:] <= 0, grad[:-1] >= 0)
  indexes = np.where(condition)[0]
  if len(indexes) >= 2:
    data = arr[indexes[-2]: indexes[-1]]
    length = np.max(data) - np.min(data)
    a = arr[indexes[-2]]
    b = arr[indexes[-1]]
    if np.abs(a - b) < tol * length:
      return indexes[-2:]
  return np.array([-1, -1])


@tools.numba_jit
def f2(arr, grad, tol):
  condition = np.logical_and(grad[:-1] * grad[1:] <= 0, grad[:-1] <= 0)
  indexes = np.where(condition)[0]
  if len(indexes) >= 2:
    data = arr[indexes[-2]: indexes[-1]]
    length = np.max(data) - np.min(data)
    a = arr[indexes[-2]]
    b = arr[indexes[-1]]
    if np.abs(a - b) < tol * length:
      return indexes[-2:]
  return np.array([-1, -1])


def find_indexes_of_limit_cycle_max(arr, tol=0.001):
  grad = np.gradient(arr)
  return f1(arr, grad, tol)


def find_indexes_of_limit_cycle_min(arr, tol=0.001):
  grad = np.gradient(arr)
  return f2(arr, grad, tol)


@tools.numba_jit
def _identity(a, b, tol=0.01):
  if np.abs(a - b) < tol:
    return True
  else:
    return False


def find_indexes_of_limit_cycle_max2(arr, tol=0.001):
  if np.ndim(arr) == 1:
    grad = np.gradient(arr)
    condition = np.logical_and(grad[:-1] * grad[1:] <= 0, grad[:-1] >= 0)
    indexes = np.where(condition)[0]
    if len(indexes) >= 2:
      data = arr[indexes[-2]: indexes[-1]]
      length = np.max(data) - np.min(data)
      if _identity(arr[indexes[-2]], arr[indexes[-1]], tol * length):
        return indexes[-2:]
    return np.array([-1, -1])

  elif np.ndim(arr) == 2:
    # The data with the shape of (axis_along_time, axis_along_neuron)
    grads = np.gradient(arr, axis=0)
    conditions = np.logical_and(grads[:-1] * grads[1:] <= 0, grads[:-1] >= 0)
    indexes = -np.ones((len(conditions), 2), dtype=int)
    for i, condition in enumerate(conditions):
      idx = np.where(condition)[0]
      if len(idx) >= 2:
        if _identity(arr[idx[-2]], arr[idx[-1]], tol):
          indexes[i] = idx[-2:]
    return indexes

  else:
    raise ValueError


def find_indexes_of_limit_cycle_min2(arr, tol=0.01):
  if np.ndim(arr) == 1:
    grad = np.gradient(arr)
    condition = np.logical_and(grad[:-1] * grad[1:] <= 0, grad[:-1] <= 0)
    indexes = np.where(condition)[0]
    if len(indexes) >= 2:
      indexes += 1
      if _identity(arr[indexes[-2]], arr[indexes[-1]], tol):
        return indexes[-2:]
    return np.array([-1, -1])

  elif np.ndim(arr) == 2:
    # The data with the shape of (axis_along_time, axis_along_neuron)
    grads = np.gradient(arr, axis=0)
    conditions = np.logical_and(grads[:-1] * grads[1:] <= 0, grads[:-1] <= 0)
    indexes = -np.ones((len(conditions), 2), dtype=int)
    for i, condition in enumerate(conditions):
      idx = np.where(condition)[0]
      if len(idx) >= 2:
        idx += 1
        if _identity(arr[idx[-2]], arr[idx[-1]], tol):
          indexes[i] = idx[-2:]
    return indexes

  else:
    raise ValueError



# -*- coding: utf-8 -*-


"""
Dynamics analysis with the aid of `SymPy <https://www.sympy.org/en/index.html>`_  symbolic inference.

This module provide basic dynamics analysis for low-dimensional dynamical systems, including

- phase plane analysis (1d or 2d systems)
- bifurcation analysis (1d or 2d systems)
- fast slow bifurcation analysis (2d or 3d systems)

"""

from .base import *
from .bifurcation import *
from .phase_plane import *
from .trajectory import *



# -*- coding: utf-8 -*-

import sys
import time
from collections import Iterable
from pprint import pprint

from brainpy import errors, math, tools

__all__ = [
  'size2len',
  'check_duration',
  'run_model',
  'check_and_format_inputs',
  'build_input_func',
  'check_and_format_monitors',
  'build_monitor_func',
]

SUPPORTED_INPUT_OPS = ['-', '+', '*', '/', '=']
SUPPORTED_INPUT_TYPE = ['fix', 'iter']


def size2len(size):
  if isinstance(size, int):
    return size
  elif isinstance(size, (tuple, list)):
    a = 1
    for b in size:
      a *= b
    return a
  else:
    raise ValueError


def check_duration(duration):
  """Check the running duration.

  Parameters
  ----------
  duration : int, list, tuple
      The running duration, it can be an int (which represents the end
      of the simulation), of a tuple/list of int (which represents the
      [start, end] / [end, start] of the simulation).

  Returns
  -------
  duration : tuple
      The tuple of running duration includes (start, end).
  """
  if isinstance(duration, (int, float)):
    start, end = 0., duration
  elif isinstance(duration, (tuple, list)):
    assert len(duration) == 2, 'Only support duration setting with the ' \
                               'format of "(start, end)" or "end".'
    start, end = duration
  else:
    raise ValueError(f'Unknown duration type: {type(duration)}. Currently, BrainPy only '
                     f'support duration specification with the format of "(start, end)" '
                     f'or "end".')

  if start > end:
    start, end = end, start
  return start, end


def get_run_length_by_duration(duration):
  start, end = check_duration(duration)
  mon_length = int((end - start) / math.get_dt())
  return mon_length


def run_model(run_func, times, report, dt=None, extra_func=None):
  """Run the model.

  The "run_func" can be the step run function of a dynamical system.

  Parameters
  ----------
  run_func : callable
      The step run function.
  times : iterable
      The model running times.
  report : float
      The percent of the total running length for each report.
  """

  # numerical integration step
  if dt is None:
    dt = math.get_dt()
  assert isinstance(dt, (int, float))

  # running function
  if extra_func is None:
    running_func = run_func
  else:
    def running_func(_t, _dt):
      extra_func(_t, _dt)
      run_func(_t, _dt)

  # simulations
  run_length = len(times)
  if report:
    t0 = time.time()
    running_func(_t=times[0], _dt=dt)
    compile_time = time.time() - t0
    print('Compilation used {:.4f} s.'.format(compile_time))

    print("Start running ...")
    report_gap = int(run_length * report)
    t0 = time.time()
    for run_idx in range(1, run_length):
      running_func(_t=times[run_idx], _dt=dt)
      if (run_idx + 1) % report_gap == 0:
        percent = (run_idx + 1) / run_length * 100
        print('Run {:.1f}% used {:.3f} s.'.format(percent, time.time() - t0))
    running_time = time.time() - t0
    print('Simulation is done in {:.3f} s.'.format(running_time))
    print()

  else:
    t0 = time.time()
    for run_idx in range(run_length):
      running_func(_t=times[run_idx], _dt=dt)
    running_time = time.time() - t0

  return running_time


def check_and_format_inputs(host, inputs):
  """Check inputs and get the formatted inputs for the given population.

  Parameters
  ----------
  host : DynamicalSystem
      The host which contains all data.
  inputs : tuple, list
      The inputs of the population.

  Returns
  -------
  formatted_inputs : tuple, list
      The formatted inputs of the population.
  """

  # 1. check inputs
  # ---------
  if inputs is None:
    inputs = []
  if not isinstance(inputs, (tuple, list)):
    raise errors.BrainPyError('"inputs" must be a tuple/list.')
  if len(inputs) > 0 and not isinstance(inputs[0], (list, tuple)):
    if isinstance(inputs[0], str):
      inputs = [inputs]
    else:
      raise errors.BrainPyError('Unknown input structure, only support inputs '
                                'with format of "(target, value, [type, operation])".')
  for one_input in inputs:
    if not 2 <= len(one_input) <= 4:
      raise errors.BrainPyError('For each target, you must specify '
                                '"(target, value, [type, operation])".')
    if len(one_input) == 3 and one_input[2] not in SUPPORTED_INPUT_TYPE:
      raise errors.BrainPyError(f'Input type only supports '
                                f'"{SUPPORTED_INPUT_TYPE}", '
                                f'not "{one_input[2]}".')
    if len(one_input) == 4 and one_input[3] not in SUPPORTED_INPUT_OPS:
      raise errors.BrainPyError(f'Input operation only supports '
                                f'"{SUPPORTED_INPUT_OPS}", '
                                f'not "{one_input[3]}".')

  # 2. get targets and attributes
  # ---------
  inputs_which_found_target = []
  inputs_not_found_target = []

  # absolute access
  nodes = host.nodes(method='absolute')
  nodes[host.name] = host
  for one_input in inputs:
    key = one_input[0]
    if not isinstance(key, str):
      raise errors.BrainPyError(f'For each input, input[0] must be a string  to '
                                f'specify variable of the target, but we got {key}.')
    splits = key.split('.')
    target = '.'.join(splits[:-1])
    key = splits[-1]
    if target == '':
      real_target = host
    else:
      if target not in nodes:
        inputs_not_found_target.append(one_input)
        continue
      real_target = nodes[target]
    if not hasattr(real_target, key):
      raise errors.BrainPyError(f'Input target key "{key}" is not defined in {real_target}.')
    inputs_which_found_target.append((real_target, key) + tuple(one_input[1:]))

  # relative access
  if len(inputs_not_found_target):
    nodes = host.nodes(method='relative')
    for one_input in inputs_not_found_target:
      splits = one_input[0].split('.')
      target, key = '.'.join(splits[:-1]), splits[-1]
      if target not in nodes:
        raise errors.BrainPyError(f'Input target "{target}" is not defined in {host}.')
      real_target = nodes[target]
      if not hasattr(real_target, key):
        raise errors.BrainPyError(f'Input target key "{key}" is not defined in {real_target}.')
      inputs_which_found_target.append((real_target, key) + tuple(one_input[1:]))

  # 3. format inputs
  # ---------
  formatted_inputs = []
  for one_input in inputs_which_found_target:
    # input value
    data_value = one_input[2]

    # input type
    if len(one_input) >= 4:
      if one_input[3] == 'iter':
        if not isinstance(data_value, Iterable):
          raise ValueError(f'Input "{data_value}" for "{one_input[0]}.{one_input[1]}" is set to be '
                           f'"iter" type, however we got the value with the type of {type(data_value)}')
      elif one_input[3] != 'fix':
        raise errors.BrainPyError(f'Only support {SUPPORTED_INPUT_TYPE} input type, but we '
                                  f'got "{one_input[3]}" in {one_input}')
      data_type = one_input[3]
    else:
      data_type = 'fix'

    # operation
    if len(one_input) == 5:
      data_op = one_input[4]
    else:
      data_op = '+'
    if data_op not in SUPPORTED_INPUT_OPS:
      raise errors.BrainPyError(f'Only support {SUPPORTED_INPUT_OPS}, while we got '
                                f'{data_op} in {one_input}')

    # final
    format_inp = one_input[:2] + (data_value, data_type, data_op)
    formatted_inputs.append(format_inp)

  return formatted_inputs


def build_input_func(inputs, show_code=False):
  input_func_name = 'input_step'
  code_scope = {'sys': sys}
  code_lines = []
  for target, key, value, type_, op in inputs:
    variable = getattr(target, key)

    # code scope
    code_scope[target.name] = target

    # code line left
    if isinstance(variable, math.Variable):
      left = f'{target.name}.{key}[...]'
    else:
      left = f'{target.name}.{key}'

    # code line right
    if type_ == 'iter':
      code_scope[f'{target.name}_input_data_of_{key}'] = iter(value)
      right = f'next({target.name}_input_data_of_{key})'
    else:
      code_scope[f'{target.name}_input_data_of_{key}'] = value
      right = f'{target.name}_input_data_of_{key}'

    # code line
    if op == '=':
      line = f'{left} = {right}'
    else:
      line = f'{left} {op}= {right}'

    code_lines.append(line)

  if len(code_lines):
    code_scope_old = {k: v for k, v in code_scope.items()}
    # function
    code, func = tools.code_lines_to_func(
      lines=code_lines,
      func_name=input_func_name,
      func_args=['_t', '_dt'],
      scope=code_scope,
      remind='Please check: \n'
             '1. whether the "iter" input is set to "fix". \n'
             '2. whether the dimensions are not match.\n')
    if show_code:
      print(code)
      print()
      pprint(code_scope_old)
      print()
  else:
    func = lambda _t, _dt: None

  return func


def check_and_format_monitors(host):
  formatted_mon_items = []

  # reshape monitors
  # ----
  all_nodes = list(host.nodes().unique().values())
  for node in all_nodes:
    node.mon.build()  # build the monitor
    for key in node.mon.item_contents.keys():
      node.mon.item_contents[key] = []  # reshape the monitor items

  # master node
  name2node = {node.name: node for node in all_nodes}
  for node in all_nodes:
    mon = node.mon
    for key, idx, interval in zip(mon.item_names, mon.item_indices, mon.item_intervals):
      # target and variable
      splits = key.split('.')
      if len(splits) == 1:
        if not hasattr(node, splits[0]):
          raise errors.BrainPyError(f'{node} does not has variable {key}.')
        target = node
        variable = splits[-1]
      else:
        if not hasattr(node, splits[0]):
          if splits[0] not in name2node:
            raise errors.BrainPyError(f'Cannot find target {key} in monitor of {node}, please check.')
          else:
            target = name2node[splits[0]]
            assert len(splits) == 2
            variable = splits[-1]
        else:
          target = node
          for s in splits[:-1]:
            try:
              target = getattr(target, s)
            except KeyError:
              raise errors.BrainPyError(f'Cannot find {key} in {node}, please check.')
          variable = splits[-1]

      # idx
      if isinstance(idx, int):
        idx = math.array([idx])

      # interval
      if interval is not None:
        if not isinstance(interval, float):
          raise errors.BrainPyError(f'"interval" must be a float (denotes time), but we got {interval}')

      # append
      formatted_mon_items.append((node, key, target, variable, idx, interval,))

  return formatted_mon_items


def build_monitor_func(monitors, show_code=False):
  """Get the monitor function according to the user's setting.

  This method will consider the following things:

  1. the monitor variable
  2. the monitor index
  3. the monitor interval

  Parameters
  ----------
  monitors : list, tuple
    The items to monitor.
  show_code : bool
      Whether show the code.
  """

  monitor_func_name = 'monitor_step'

  code_lines = []
  code_scope = {'sys': sys}

  for node, key, target, variable, idx, interval in monitors:
    code_scope[node.name] = node
    code_scope[target.name] = target

    # get data
    data = target
    for k in variable.split('.'):
      data = getattr(data, k)

    # get the data key in the host

    if isinstance(data, (int, float)):
      if idx is not None:
        raise errors.BrainPyError(f'"{key}" is a scalar, cannot define the slice index "{idx}"')
      key_in_host = f'{target.name}.{variable}'
    elif math.ndim(data) == 1:
      if hasattr(data, 'value'):
        key_in_host = f'{target.name}.{variable}.value.copy()'
      else:
        key_in_host = f'{target.name}.{variable}.copy()'
    else:
      if hasattr(data, 'value'):
        key_in_host = f'{target.name}.{variable}.value.flatten().copy()'
      else:
        key_in_host = f'{target.name}.{variable}.flatten().copy()'

    # format the monitor index
    if idx is None:
      right = key_in_host
    else:
      if hasattr(idx, 'value'):
        idx = idx.value
      right = f'{key_in_host}[{node.name}_mon_{key.replace(".", "_")}_idx]'
      code_scope[f'{node.name}_mon_{key.replace(".", "_")}_idx'] = idx

    # format the monitor lines according to the time interval
    if interval is None:
      code_lines.append(f'{node.name}.mon.item_contents["{key}"].append({right})')
    else:
      code_scope[f'{node.name}_mon_{key.replace(".", "_")}_next_time'] = interval
      # code_scope[f'{node.name}_mon_{key.replace(".", "_")}_interval'] = interval
      code_lines.extend([f'global {node.name}_mon_{key.replace(".", "_")}_next_time',
                         f'if _t >= {node.name}_mon_{key.replace(".", "_")}_next_time:',
                         f'  {node.name}.mon.item_contents["{key}"].append({right})',
                         f'  {node.name}.mon.item_contents["{key}.t"].append(_t)',
                         f'  {node.name}_mon_{key.replace(".", "_")}_next_time += {interval}'])

  if len(code_lines):
    code_scope_old = {k: v for k, v in code_scope.items()}
    # function
    code, func = tools.code_lines_to_func(lines=code_lines,
                                          func_name=monitor_func_name,
                                          func_args=['_t', '_dt'],
                                          scope=code_scope)
    if show_code:
      print(code)
      print()
      pprint(code_scope_old)
      print()
  else:
    func = lambda _t, _dt: None

  return func



# -*- coding: utf-8 -*-

r"""This module provides exponential integrators for ODEs.

Exponential integrators are a large class of methods from numerical analysis is based on
the exact integration of the linear part of the initial value problem. Because the linear
part is integrated exactly, this can help to mitigate the stiffness of a differential
equation.

We consider initial value problems of the form,

.. math:: u'(t)=f(u(t)),\qquad u(t_{0})=u_{0},

which can be decomposed of

.. math:: u'(t)=Lu(t)+N(u(t)),\qquad u(t_{0})=u_{0},

where :math:`L={\frac {\partial f}{\partial u}}` (the Jacobian of f) is composed of
linear terms, and :math:`N=f(u)-Lu` is composed of the non-linear terms.

This procedure enjoys the advantage, in each step, that
:math:`{\frac {\partial N_{n}}{\partial u}}(u_{n})=0`.
This considerably simplifies the derivation of the order conditions and improves the
stability when integrating the nonlinearity :math:`N(u(t))`.

Exact integration of this problem from time 0 to a later time :math:`t` can be performed
using `matrix exponentials <https://en.wikipedia.org/wiki/Matrix_exponential>`_ to define
an integral equation for the exact solution:

.. math:: u(t)=e^{Lt}u_{0}+\int _{0}^{t}e^{L(t-\tau )}N\left(t+\tau, u\left(\tau \right)\right)\,d\tau .

This representation of the exact solution is also called as *variation-of-constant formula*.
In the case of :math:`N\equiv 0`, this formulation is the exact solution to the linear
differential equation.


**Exponential Rosenbrock methods**

Exponential Rosenbrock methods were shown to be very efficient in solving large systems
of stiff ODEs. Applying the variation-of-constants formula gives the exact solution at
time :math:`t_{n+1}` with the numerical solution :math:`u_n` as

.. math::
    u(t_{n+1})=e^{h_{n}L}u(t_{n})+\int _{0}^{h_{n}}e^{(h_{n}-\tau )L}N(t_n+\tau, u(t_{n}+\tau ))d\tau .
    :label: discrete-variation-of-constants-formula

where :math:`h_n=t_{n+1}-t_n`.

The idea now is to approximate the integral in :eq:`discrete-variation-of-constants-formula`
by some quadrature rule with nodes :math:`c_{i}` and weights :math:`b_{i}(h_{n}L)`
(:math:`1\leq i\leq s`). This yields the following class of *s-stage* explicit exponential
Rosenbrock methods:

.. math::
    \begin{align}
    U_{ni}=&e^{c_{i}h_{n}L}u_n+h_{n}\sum_{j=1}^{i-1}a_{ij}(h_{n}L)N(U_{nj}),  \\
    u_{n+1}=&e^{h_{n}L}u_n+h_{n}\sum_{i=1}^{s}b_{i}(h_{n}L)N(U_{ni})
    \end{align}

where :math:`U_{ni}\approx u(t_{n}+c_{i}h_{n})`.

The coefficients :math:`a_{ij}(z),b_{i}(z)` are usually chosen as linear combinations of
the entire functions :math:`\varphi _{k}(c_{i}z),\varphi _{k}(z)`, respectively, where

.. math::
    \begin{align}
    \varphi _{k}(z)=&\int _{0}^{1}e^{(1-\theta )z}{\frac {\theta ^{k-1}}{(k-1)!}}d\theta ,\quad k\geq 1, \\
    \varphi _{0}(z)=&e^{z},\\
    \varphi _{k+1}(z)=&{\frac {\varphi_{k}(z)-\varphi _{k}(0)}{z}},\ k\geq 0.
    \end{align}

By introducing the difference :math:`D_{ni}=N(U_{ni})-N(u_{n})`, they can be reformulated
in a more efficient way for implementation as

.. math::
    \begin{align}
    U_{ni}=&u_{n}+c_{i}h_{n}\varphi _{1}(c_{i}h_{n}L)f(u_{n})+h_{n}\sum _{j=2}^{i-1}a_{ij}(h_{n}L)D_{nj}, \\
    u_{n+1}=&u_{n}+h_{n}\varphi _{1}(h_{n}L)f(u_{n})+h_{n}\sum _{i=2}^{s}b_{i}(h_{n}L)D_{ni}.
    \end{align}

where :math:`\varphi_{1}(z)=\frac{e^z-1}{z}`.

In order to implement this scheme with adaptive step size, one can consider, for the purpose
of local error estimation, the following embedded methods

.. math:: {\bar {u}}_{n+1}=u_{n}+h_{n}\varphi _{1}(h_{n}L)f(u_{n})+h_{n}\sum _{i=2}^{s}{\bar {b}}_{i}(h_{n}L)D_{ni},

which use the same stages :math:`U_{ni}` but with weights :math:`{\bar {b}}_{i}`.

For convenience, the coefficients of the explicit exponential Rosenbrock methods together
with their embedded methods can be represented by using the so-called reduced Butcher
tableau as follows:

.. math::
    \begin{array}{c|ccccc}
    c_{2} & & & & & \\
    c_{3} & a_{32} & & & & \\
    \vdots & \vdots & & \ddots & & \\
    c_{s} & a_{s 2} & a_{s 3} & \cdots & a_{s, s-1} \\
    \hline & b_{2} & b_{3} & \cdots & b_{s-1} & b_{s} \\
    & \bar{b}_{2} & \bar{b}_{3} & \cdots & \bar{b}_{s-1} & \bar{b}_{s}
    \end{array}

.. [1] https://en.wikipedia.org/wiki/Exponential_integrator
.. [2] Hochbruck, M., & Ostermann, A. (2010). Exponential integrators. Acta Numerica, 19, 209-286.
"""

from brainpy import math
from brainpy.integrators import constants
from brainpy.integrators.ode.wrapper import exp_euler_wrapper

__all__ = [
  'exponential_euler',
  'exponential_midpoint',
]


def exponential_euler(f=None, show_code=None, dt=None, var_type=None):
  r"""The exponential Euler method for ODEs.

  The simplest exponential Rosenbrock method is the exponential
  Rosenbrock–Euler scheme, which has order 2.

  For an ODE equation of the form

  .. math::

      u^{\prime}=f(u), \quad u(0)=u_{0}

  its schema is given by

  .. math::

      u_{n+1}= u_{n}+h \varphi(hL) f (u_{n})

  where :math:`L=f^{\prime}(u_{n})` and :math:`\varphi(z)=\frac{e^{z}-1}{z}`.

  For a linear ODE system: :math:`u^{\prime} = Ay + B`,
  the above equation is equal to :math:`u_{n+1}= u_{n}e^{hA}-B/A(1-e^{hA})`,
  which is the exact solution for this ODE system.

  Parameters
  ----------
  f : function
    The derivative function.
  dt : optional, float
    The numerical precision.
  var_type : optional, str
    The variable type.
  show_code : bool
    Whether show the code.

  Returns
  -------
  func : function
      The one-step numerical integrator function.
  """

  dt = math.get_dt() if dt is None else dt
  show_code = False if show_code is None else show_code
  var_type = constants.SCALAR_VAR if var_type is None else var_type

  if f is None:
    return lambda f: exp_euler_wrapper(f,
                                       show_code=show_code,
                                       dt=dt,
                                       var_type=var_type)
  else:
    return exp_euler_wrapper(f,
                             show_code=show_code,
                             dt=dt,
                             var_type=var_type)


def exponential_midpoint(f=None, show_code=None, dt=None, var_type=None):
  r"""The exponential midpoint method for ODEs.

  Borgers and Nectow [1]_ [2]_ considered an explicit second-order method, called
  the exponential midpoint method, which is given by:

  .. math::
      \begin{align}
      u_{n+1/2} = &e^{1/2hL(u_n)}u_n + \frac{e^{1/2hL}-1}{1/2hL(u_n)} \frac{1}{2}hN(U_n) \\
      u_{n+1} = & e^{hL(u_{n+1/2})}u_n + \frac{e^{hL(u_{n+1/2})}-1}{hL(u_{n+1/2})}hN(u_{n+1/2})
      \end{align}

  The second line is similar to exponential Euler, except :math:`L` and :math:`N` are
  evaluated at the approximate midpoint :math:`u_{n+1/2}`, obtained by first taking
  a half-step of exponential Euler. Note that each nonlinear function is evaluated
  twice, first at :math:`u_{n}` and again at :math:`u_{n+1/2}`, so this method is
  twice as expensive per step as an Euler-type method.

  Parameters
  ----------
  f : function
    The derivative function.
  dt : optional, float
    The numerical precision.
  var_type : optional, str
    The variable type.
  show_code : bool
    Whether show the code.

  Returns
  -------
  func : function
      The one-step numerical integrator function.

  References
  ----------
  .. [1] Borgers, C., & Nectow, A. R. (2013). Exponential Time Differencing for
         Hodgkin--Huxley-like ODEs. SIAM Journal on Scientific Computing, 35(3),
         B623-B643.
  .. [2] Chen, Zhengdao, Baranidharan Raman, and Ari Stern. "Structure-Preserving
         Numerical Integrators for Hodgkin--Huxley-Type Systems." SIAM Journal on
         Scientific Computing 42, no. 1 (2020): B273-B298.
  """
  raise NotImplementedError



# -*- coding: utf-8 -*-

import numpy.fft

__all__ = [
  "fft", "fft2", "fftfreq", "fftn", "fftshift", "hfft",
  "ifft", "ifft2", "ifftn", "ifftshift", "ihfft", "irfft",
  "irfft2", "irfftn", "rfft", "rfft2", "rfftfreq", "rfftn"
]

fft = numpy.fft.fft
fft2 = numpy.fft.fft2
fftfreq = numpy.fft.fftfreq
fftn = numpy.fft.fftn
fftshift = numpy.fft.fftshift
hfft = numpy.fft.hfft
ifft = numpy.fft.ifft
ifft2 = numpy.fft.ifft2
ifftn = numpy.fft.ifftn
ifftshift = numpy.fft.ifftshift
ihfft = numpy.fft.ihfft
irfft = numpy.fft.irfft
irfft2 = numpy.fft.irfft2
irfftn = numpy.fft.irfftn
rfft = numpy.fft.rfft
rfft2 = numpy.fft.rfft2
rfftfreq = numpy.fft.rfftfreq
rfftn = numpy.fft.rfftn



# -*- coding: utf-8 -*-


r"""This module provides commonly used activation functions.

Activation functions are a critical part of the design of a neural network.
The choice of activation function in the hidden layer will control how well
the network model learns the training dataset. The choice of activation
function in the output layer will define the type of predictions the model
can make.

"""

import operator

import numpy as np
import jax
import jax.numpy
import jax.scipy
import jax.core

from brainpy.math.jax import ops
from brainpy.math.jax.jaxarray import JaxArray

__all__ = [
  'celu',
  'elu',
  'gelu',
  'glu',
  'hard_tanh',
  'hard_sigmoid',
  'hard_silu',
  'hard_swish',
  'leaky_relu',
  'log_sigmoid',
  'log_softmax',
  'one_hot',
  'normalize',
  'relu',
  'relu6',
  'sigmoid',
  'soft_sign',
  'softmax',
  'softplus',
  'silu',
  'swish',
  'selu',

  'tanh',

]


def get(activation):
  global_vars = globals()

  if activation not in global_vars:
    raise ValueError(f'Unknown activation function: {activation}, \nwe only support: '
                     f'{[k for k, v in global_vars.items() if not k.startswith("_") and callable(v)]}')
  return global_vars[activation]


def celu(x, alpha=1.0):
  r"""Continuously-differentiable exponential linear unit activation.

  Computes the element-wise function:

  .. math::
    \mathrm{celu}(x) = \begin{cases}
      x, & x > 0\\
      \alpha \left(\exp(\frac{x}{\alpha}) - 1\right), & x \le 0
    \end{cases}

  For more information, see
  `Continuously Differentiable Exponential Linear Units
  <https://arxiv.org/pdf/1704.07483.pdf>`_.

  Parameters
  ----------
  x : ndarray, jnp.ndarray
    The input array.
  alpha : ndarray, float
    The default is 1.0.
  """
  x = x.value if isinstance(x, JaxArray) else x
  alpha = alpha.value if isinstance(alpha, JaxArray) else alpha
  return JaxArray(jax.numpy.where(x > 0, x, alpha * jax.numpy.expm1(x / alpha)))


def elu(x, alpha=1.0):
  r"""Exponential linear unit activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{elu}(x) = \begin{cases}
      x, & x > 0\\
      \alpha \left(\exp(x) - 1\right), & x \le 0
    \end{cases}

  Args:
    x : input array
    alpha : scalar or array of alpha values (default: 1.0)
  """
  x = x.value if isinstance(x, JaxArray) else x
  alpha = alpha.value if isinstance(alpha, JaxArray) else alpha
  safe_x = jax.numpy.where(x > 0, 0., x)
  return JaxArray(jax.numpy.where(x > 0, x, alpha * jax.numpy.expm1(safe_x)))


def gelu(x, approximate=True):
  r"""Gaussian error linear unit activation function.

  If ``approximate=False``, computes the element-wise function:

  .. math::
    \mathrm{gelu}(x) = \frac{x}{2} \left(1 + \mathrm{erf} \left(
      \frac{x}{\sqrt{2}} \right) \right)

  If ``approximate=True``, uses the approximate formulation of GELU:

  .. math::
    \mathrm{gelu}(x) = \frac{x}{2} \left(1 + \mathrm{tanh} \left(
      \sqrt{\frac{2}{\pi}} \left(x + 0.044715 x^3 \right) \right) \right)

  For more information, see `Gaussian Error Linear Units (GELUs)
  <https://arxiv.org/abs/1606.08415>`_, section 2.

  Args:
    x : input array
    approximate: whether to use the approximate or exact formulation.
  """
  x = x.value if isinstance(x, JaxArray) else x
  if approximate:
    sqrt_2_over_pi = np.sqrt(2 / np.pi).astype(x.dtype)
    cdf = 0.5 * (1.0 + jax.numpy.tanh(sqrt_2_over_pi * (x + 0.044715 * (x ** 3))))
    y = x * cdf
  else:
    y = jax.numpy.array(x * (jax.lax.erf(x / np.sqrt(2)) + 1) / 2, dtype=x.dtype)
  return JaxArray(y)


def glu(x, axis=-1):
  r"""Gated linear unit activation function.

  Args:
    x : input array
    axis: the axis along which the split should be computed (default: -1)
  """
  size = x.shape[axis]
  assert size % 2 == 0, "axis size must be divisible by 2"
  x = x.value if isinstance(x, JaxArray) else x
  x1, x2 = jax.numpy.split(x, 2, axis)
  return JaxArray(x1 * sigmoid(x2))


def hard_tanh(x):
  r"""Hard :math:`\mathrm{tanh}` activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{hard\_tanh}(x) = \begin{cases}
      -1, & x < -1\\
      x, & -1 \le x \le 1\\
      1, & 1 < x
    \end{cases}

  Args:
    x : input array
  """
  x = x.value if isinstance(x, JaxArray) else x
  return JaxArray(jax.numpy.where(x > 1, 1, jax.numpy.where(x < -1, -1, x)))


def hard_sigmoid(x):
  r"""Hard Sigmoid activation function.

  Computes the element-wise function

  .. math::
    \mathrm{hard\_sigmoid}(x) = \frac{\mathrm{relu6}(x + 3)}{6}

  Args:
    x : input array
  """
  return relu6(x + 3.) / 6.


def hard_silu(x):
  r"""Hard SiLU activation function

  Computes the element-wise function

  .. math::
    \mathrm{hard\_silu}(x) = x \cdot \mathrm{hard\_sigmoid}(x)

  Args:
    x : input array
  """
  return x * hard_sigmoid(x)


hard_swish = hard_silu


def leaky_relu(x, negative_slope=1e-2):
  r"""Leaky rectified linear unit activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{leaky\_relu}(x) = \begin{cases}
      x, & x \ge 0\\
      \alpha x, & x < 0
    \end{cases}

  where :math:`\alpha` = :code:`negative_slope`.

  Args:
    x : input array
    negative_slope  or scalar specifying the negative slope (default: 0.01)
  """
  x = x.value if isinstance(x, JaxArray) else x
  return JaxArray(jax.numpy.where(x >= 0, x, negative_slope * x))


def softplus(x):
  r"""Softplus activation function.

  Computes the element-wise function

  .. math::
    \mathrm{softplus}(x) = \log(1 + e^x)

  Args:
    x : input array
  """
  x = x.value if isinstance(x, JaxArray) else x
  return JaxArray(jax.numpy.logaddexp(x, 0))


def log_sigmoid(x):
  r"""Log-sigmoid activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{log\_sigmoid}(x) = \log(\mathrm{sigmoid}(x)) = -\log(1 + e^{-x})

  Args:
    x : input array
  """
  return -softplus(-x)


def log_softmax(x, axis=-1):
  r"""Log-Softmax function.

  Computes the logarithm of the :code:`softmax` function, which rescales
  elements to the range :math:`[-\infty, 0)`.

  .. math ::
    \mathrm{log\_softmax}(x) = \log \left( \frac{\exp(x_i)}{\sum_j \exp(x_j)}
    \right)

  Args:
    x : input array
    axis: the axis or axes along which the :code:`log_softmax` should be
      computed. Either an integer or a tuple of integers.
  """
  x = x.value if isinstance(x, JaxArray) else x
  shifted = x - jax.lax.stop_gradient(x.max(axis, keepdims=True))
  return JaxArray(shifted - jax.numpy.log(jax.numpy.sum(jax.numpy.exp(shifted), axis, keepdims=True)))


def _canonicalize_axis(axis, num_dims) -> int:
  """Canonicalize an axis in [-num_dims, num_dims) to [0, num_dims)."""
  axis = operator.index(axis)
  if not -num_dims <= axis < num_dims:
    raise ValueError(
      "axis {} is out of bounds for array of dimension {}".format(
        axis, num_dims))
  if axis < 0:
    axis = axis + num_dims
  return axis


def one_hot(x, num_classes, *, dtype=None, axis=-1):
  r"""One-hot encodes the given indicies.

  Each index in the input ``x`` is encoded as a vector of zeros of length
  ``num_classes`` with the element at ``index`` set to one::

    >>> jax.nn.one_hot(ops.array([0, 1, 2]), 3)
    DeviceArray([[1., 0., 0.],
                  [0., 1., 0.],
                  [0., 0., 1.]], dtype=float32)

  Indicies outside the range [0, num_classes) will be encoded as zeros::

    >>> jax.nn.one_hot(ops.array([-1, 3]), 3)
    DeviceArray([[0., 0., 0.],
                 [0., 0., 0.]], dtype=float32)

  Args:
    x: A tensor of indices.
    num_classes: Number of classes in the one-hot dimension.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    axis: the axis or axes along which the function should be
      computed.
  """
  num_classes = jax.core.concrete_or_error(
    int, num_classes, "The error arose in jax.nn.one_hot argument `num_classes`.")
  dtype = jax.dtypes.canonicalize_dtype(jax.numpy.float64 if dtype is None else dtype)
  x = jax.numpy.asarray(x)
  try:
    output_pos_axis = _canonicalize_axis(axis, x.ndim + 1)
  except TypeError:
    axis_size = jax.lax.psum(1, axis)
    if num_classes != axis_size:
      raise ValueError(f"Expected num_classes to match the size of axis {axis}, "
                       f"but {num_classes} != {axis_size}") from None
    axis_idx = jax.lax.axis_index(axis)
    return jax.numpy.asarray(x == axis_idx, dtype=dtype)
  axis = operator.index(axis)
  lhs = jax.lax.expand_dims(x, (axis,))
  rhs_shape = [1] * x.ndim
  rhs_shape.insert(output_pos_axis, num_classes)
  rhs = jax.lax.broadcast_in_dim(jax.numpy.arange(num_classes, dtype=x.dtype),
                                 rhs_shape, (output_pos_axis,))
  return JaxArray(jax.numpy.asarray(lhs == rhs, dtype=dtype))


def normalize(x, axis=-1, mean=None, variance=None, epsilon=1e-5):
  """Normalizes an array by subtracting mean and dividing by sqrt(var)."""
  x = x.value if isinstance(x, JaxArray) else x
  if mean is None:
    mean = jax.numpy.mean(x, axis, keepdims=True)
  if variance is None:
    # this definition is traditionally seen as less accurate than jnp.var's
    # mean((x - mean(x))**2) but may be faster and even, given typical
    # activation distributions and low-precision arithmetic, more accurate
    # when used in neural network normalization layers
    variance = jax.numpy.mean(jax.numpy.square(x), axis, keepdims=True) - jax.numpy.square(mean)
  y = (x - mean) * jax.lax.rsqrt(variance + epsilon)
  return JaxArray(y)


def relu(x):
  x = x.value if isinstance(x, JaxArray) else x
  return JaxArray(jax.nn.relu(x))


def relu6(x):
  r"""Rectified Linear Unit 6 activation function.

  Computes the element-wise function

  .. math::
    \mathrm{relu6}(x) = \min(\max(x, 0), 6)

  Args:
    x : input array
  """
  x = x.value if isinstance(x, JaxArray) else x
  return JaxArray(jax.numpy.minimum(jax.numpy.maximum(x, 0), 6.))


def sigmoid(x):
  r"""Sigmoid activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{sigmoid}(x) = \frac{1}{1 + e^{-x}}

  Args:
    x : input array
  """
  x = x.value if isinstance(x, JaxArray) else x
  return JaxArray(jax.scipy.special.expit(x))


def soft_sign(x):
  r"""Soft-sign activation function.

  Computes the element-wise function

  .. math::
    \mathrm{soft\_sign}(x) = \frac{x}{|x| + 1}

  Args:
    x : input array
  """
  x = x.value if isinstance(x, JaxArray) else x
  return JaxArray(x / (jax.numpy.abs(x) + 1))


def softmax(x, axis=-1):
  r"""Softmax function.

  Computes the function which rescales elements to the range :math:`[0, 1]`
  such that the elements along :code:`axis` sum to :math:`1`.

  .. math ::
    \mathrm{softmax}(x) = \frac{\exp(x_i)}{\sum_j \exp(x_j)}

  Args:
    x : input array
    axis: the axis or axes along which the softmax should be computed. The
      softmax output summed across these dimensions should sum to :math:`1`.
      Either an integer or a tuple of integers.
  """
  x = x.value if isinstance(x, JaxArray) else x
  unnormalized = jax.numpy.exp(x - jax.lax.stop_gradient(x.max(axis, keepdims=True)))
  return JaxArray(unnormalized / unnormalized.sum(axis, keepdims=True))


def silu(x):
  r"""SiLU activation function.

  Computes the element-wise function:

  .. math::
    \mathrm{silu}(x) = x \cdot \mathrm{sigmoid}(x) = \frac{x}{1 + e^{-x}}

  Args:
    x : input array
  """
  x = x.value if isinstance(x, JaxArray) else x
  return JaxArray(x * sigmoid(x))


swish = silu


def selu(x):
  r"""Scaled exponential linear unit activation.

  Computes the element-wise function:

  .. math::
    \mathrm{selu}(x) = \lambda \begin{cases}
      x, & x > 0\\
      \alpha e^x - \alpha, & x \le 0
    \end{cases}

  where :math:`\lambda = 1.0507009873554804934193349852946` and
  :math:`\alpha = 1.6732632423543772848170429916717`.

  For more information, see
  `Self-Normalizing Neural Networks
  <https://papers.nips.cc/paper/6698-self-normalizing-neural-networks.pdf>`_.

  Args:
    x : input array
  """
  alpha = 1.6732632423543772848170429916717
  scale = 1.0507009873554804934193349852946
  x = x.value if isinstance(x, JaxArray) else x
  safe_x = jax.numpy.where(x > 0, 0., x)
  return JaxArray(scale * jax.numpy.where(x > 0, x, alpha * jax.numpy.expm1(safe_x)))


def tanh(x):
  x = x.value if isinstance(x, JaxArray) else x
  return JaxArray(jax.lax.tanh(x))


# -*- coding: utf-8 -*-

"""
This module provides analysis tools for ordinary differential equations.

The first part is the analysis package with SymPy symbolic inference.

The second part is the analysis package with numerical continuation methods.

Moreover, we provide several useful functions which may help your dynamical
system analysis, like:

>>> get_1d_stability_types()
['saddle node', 'stable point', 'unstable point']

"""

from . import numeric
from . import symbolic
from . import continuation
from . import solver
from . import stability
from .solver import *
from .stability import *



# -*- coding: utf-8 -*-

import abc
from brainpy import math

from brainpy.simulation.brainobjects.base import DynamicalSystem
from brainpy.simulation.brainobjects.neuron import NeuGroup
from brainpy.simulation.initialize import XavierNormal, ZeroInit, Uniform, Orthogonal

__all__ = [
  'RNNCore',
  'VanillaRNN',
  'GRU',
  'LSTM',
]


class RNNCore(NeuGroup):
  def __init__(self, num_hidden, num_input, **kwargs):
    super(RNNCore, self).__init__(size=num_hidden, **kwargs)
    assert isinstance(num_hidden, int)
    assert isinstance(num_input, int)
    self.num_hidden = num_hidden
    self.num_input = num_input

  @abc.abstractmethod
  def init(self, num_batch=1, **kwargs):
    pass

  @abc.abstractmethod
  def update(self, x, **kwargs):
    pass


class VanillaRNN(RNNCore):
  r"""Basic fully-connected RNN core.

  Given :math:`x_t` and the previous hidden state :math:`h_{t-1}` the
  core computes

  .. math::

     h_t = \mathrm{ReLU}(w_i x_t + b_i + w_h h_{t-1} + b_h)

  The output is equal to the new state, :math:`h_t`.
  """
  target_backend = 'jax'

  def __init__(self, num_hidden, num_input, w_init=XavierNormal(), b_init=ZeroInit(),
               h_init=Uniform(), **kwargs):
    super(VanillaRNN, self).__init__(num_hidden, num_input, **kwargs)

    # parameters
    self.h_init = h_init

    # weights
    self.w_ir = math.TrainVar(w_init((num_input, num_hidden)))
    self.w_rr = math.TrainVar(w_init((num_hidden, num_hidden)))
    self.b = math.TrainVar(b_init((num_hidden,)))

  def update(self, x, **kwargs):
    self.h[:] = math.relu(x @ self.w_ir + self.h @ self.w_rr + self.b)
    return self.h

  def init(self, num_batch=1, **kwargs):
    self.num_batch = num_batch
    self.h = math.Variable(self.h_init((num_batch, self.num_hidden)))


class GRU(RNNCore):
  r"""Gated Recurrent Unit.

  The implementation is based on (Chung, et al., 2014) [1]_ with biases.

  Given :math:`x_t` and the previous state :math:`h_{t-1}` the core computes

  .. math::

     \begin{array}{ll}
     z_t &= \sigma(W_{iz} x_t + W_{hz} h_{t-1} + b_z) \\
     r_t &= \sigma(W_{ir} x_t + W_{hr} h_{t-1} + b_r) \\
     a_t &= \tanh(W_{ia} x_t + W_{ha} (r_t \bigodot h_{t-1}) + b_a) \\
     h_t &= (1 - z_t) \bigodot h_{t-1} + z_t \bigodot a_t
     \end{array}

  where :math:`z_t` and :math:`r_t` are reset and update gates.

  The output is equal to the new hidden state, :math:`h_t`.

  Warning: Backwards compatibility of GRU weights is currently unsupported.

  References
  ----------
  .. [1] Chung, J., Gulcehre, C., Cho, K. and Bengio, Y., 2014. Empirical
         evaluation of gated recurrent neural networks on sequence modeling.
         arXiv preprint arXiv:1412.3555.
  """

  def __init__(self, num_hidden, num_input, wx_init=Orthogonal(),
               wh_init=Orthogonal(), b_init=ZeroInit(), h_init=ZeroInit(), **kwargs):
    super(GRU, self).__init__(num_hidden, num_input**kwargs)

    # parameters
    self.h_init = h_init

    # weights
    self.w_iz = math.TrainVar(wx_init((num_input, num_hidden)))
    self.w_ir = math.TrainVar(wx_init((num_input, num_hidden)))
    self.w_ia = math.TrainVar(wx_init((num_input, num_hidden)))
    self.w_hz = math.TrainVar(wh_init((num_hidden, num_hidden)))
    self.w_hr = math.TrainVar(wh_init((num_hidden, num_hidden)))
    self.w_ha = math.TrainVar(wh_init((num_hidden, num_hidden)))
    self.bz = math.TrainVar(b_init((num_hidden,)))
    self.br = math.TrainVar(b_init((num_hidden,)))
    self.ba = math.TrainVar(b_init((num_hidden,)))

  def update(self, x, **kwargs):
    z = math.sigmoid(x @ self.w_iz + self.h @ self.w_hz + self.bz)
    r = math.sigmoid(x @ self.w_ir + self.h @ self.w_hr + self.br)
    a = math.tanh(x @ self.w_ia + (r * self.h) @ self.w_ha + self.ba)
    self.h[:] = (1 - z) * self.h + z * a
    return self.h

  def init(self, num_batch=1, **kwargs):
    self.num_batch = num_batch
    self.h = math.Variable(self.h_init((num_batch, self.num_hidden)))


class LSTM(RNNCore):
  r"""Long short-term memory (LSTM) RNN core.

  The implementation is based on (zaremba, et al., 2014) [1]_. Given
  :math:`x_t` and the previous state :math:`(h_{t-1}, c_{t-1})` the core
  computes

  .. math::

     \begin{array}{ll}
     i_t = \sigma(W_{ii} x_t + W_{hi} h_{t-1} + b_i) \\
     f_t = \sigma(W_{if} x_t + W_{hf} h_{t-1} + b_f) \\
     g_t = \tanh(W_{ig} x_t + W_{hg} h_{t-1} + b_g) \\
     o_t = \sigma(W_{io} x_t + W_{ho} h_{t-1} + b_o) \\
     c_t = f_t c_{t-1} + i_t g_t \\
     h_t = o_t \tanh(c_t)
     \end{array}

  where :math:`i_t`, :math:`f_t`, :math:`o_t` are input, forget and
  output gate activations, and :math:`g_t` is a vector of cell updates.

  The output is equal to the new hidden, :math:`h_t`.

  Notes
  -----

  Forget gate initialization: Following (Jozefowicz, et al., 2015) [2]_ we add 1.0
  to :math:`b_f` after initialization in order to reduce the scale of forgetting in
  the beginning of the training.

  References
  ----------

  .. [1] Zaremba, Wojciech, Ilya Sutskever, and Oriol Vinyals. "Recurrent neural
         network regularization." arXiv preprint arXiv:1409.2329 (2014).
  .. [2] Jozefowicz, Rafal, Wojciech Zaremba, and Ilya Sutskever. "An empirical
         exploration of recurrent network architectures." In International conference
         on machine learning, pp. 2342-2350. PMLR, 2015.
  """

  def __init__(self, num_hidden, num_input,
               w_init=Orthogonal(), b_init=ZeroInit(),
               h_init=ZeroInit(), **kwargs):
    super(LSTM, self).__init__(num_hidden, num_input, **kwargs)

    # parameters
    self.h_init = h_init

    # weights
    self.w = math.TrainVar(w_init((num_input + num_hidden, num_hidden * 4)))
    self.b = math.TrainVar(b_init((num_hidden * 4,)))

  def update(self, x, **kwargs):
    xh = math.concatenate([x, self.h], axis=-1)
    gated = xh @ self.w + self.b
    i, g, f, o = math.split(gated, indices_or_sections=4, axis=-1)
    c = math.sigmoid(f + 1.) * self.c + math.sigmoid(i) * math.tanh(g)
    h = math.sigmoid(o) * math.tanh(c)
    self.h[:] = h
    self.c[:] = c
    return h

  def init(self, num_batch=1, **kwargs):
    self.num_batch = num_batch
    self.h = math.Variable(self.h_init((num_batch, self.num_hidden)))
    self.c = math.Variable(self.h_init((num_batch, self.num_hidden)))


class ConvLSTM(RNNCore):
  pass



# -*- coding: utf-8 -*-


import jax
import jax.numpy as jnp

from brainpy.math.jax import Variable


def test_varaible1():
  @jax.jit
  def try_variable1(a, b):
    return a + b


  va = Variable(jnp.zeros(10), replicate=lambda *args: jnp.zeros(10))
  vb = Variable(jnp.ones(10), replicate=lambda *args: jnp.ones(10))
  va.duplicate(10)
  vb.duplicate(10)
  print(try_variable1(va, vb))




# -*- coding: utf-8 -*-

from brainpy.simulation.initialize import XavierNormal, ZeroInit
from brainpy.simulation.layers.activation import Activation
from brainpy.simulation.layers.dense import Dense
from brainpy.simulation.layers.sequential import Sequential

__all__ = [
  'MLP'
]


class MLP(Sequential):
  """Multi-layer perceptron.
  """

  def __init__(self, layer_sizes, activation='relu', w_init=XavierNormal(), b_init=ZeroInit(), name=None):
    assert len(layer_sizes) >= 2
    name = self.unique_name(name)
    layers = []
    for i in range(1, len(layer_sizes)):
      layers.append(Dense(num_input=layer_sizes[i - 1], num_hidden=layer_sizes[i],
                          w_init=w_init, b_init=b_init, name=f'{name}_l{i}'))
      layers.append(Activation(activation))
    super(MLP, self).__init__(*layers, name=name)



# -*- coding: utf-8 -*-

import brainpy as bp
from brainpy.base import Base

bp.math.use_backend('jax')


def test_grad1():
  class Test(Base):
    def __init__(self):
      super(Test, self).__init__()

      self.a = bp.math.TrainVar(bp.math.ones(10))
      self.b = bp.math.TrainVar(bp.math.random.randn(10))
      self.c = bp.math.TrainVar(bp.math.random.uniform(size=10))

    def __call__(self):
      return bp.math.sum(self.a + self.b + self.c)

  o2 = Test()
  o2_grad = bp.math.grad(o2, o2.vars())
  grads = o2_grad()

  print('test_grad1:')
  print(grads)
  for g in grads.values():
    assert (g == 1.).all()


def test_grad2():
  class Test(Base):
    def __init__(self):
      super(Test, self).__init__()

      self.a = bp.math.TrainVar(bp.math.ones(10))
      self.b = bp.math.TrainVar(bp.math.random.randn(10))
      self.c = bp.math.TrainVar(bp.math.random.uniform(size=10))

    def __call__(self, d):
      return bp.math.sum(self.a + self.b + self.c + d)

  o2 = Test()
  o2_grad = bp.math.grad(o2, o2.vars())

  res = o2_grad(bp.math.random.random(10))
  print('test_grad2:')
  print(res)
  for g in res.values():
    assert (g == 1.).all()


def test_grad3():
  class Test(Base):
    def __init__(self):
      super(Test, self).__init__()

      self.a = bp.math.TrainVar(bp.math.ones(10))
      self.b = bp.math.TrainVar(bp.math.random.randn(10))
      self.c = bp.math.TrainVar(bp.math.random.uniform(size=10))

    def __call__(self, d):
      return bp.math.sum(self.a + self.b + self.c + 2 * d)

  o2 = Test()
  o2_grad = bp.math.grad(o2, o2.vars(), argnums=0)

  res = o2_grad(bp.math.random.random(10))
  print('test_grad3:')
  print(res)
  for g in res[1].values():
    assert (g == 1.).all()
  assert (res[0] == 2.).all()


def test_grad_aux1():
  class Test(bp.dnn.Module):

    def __init__(self):
      super(Test, self).__init__()

      self.a = bp.math.TrainVar(bp.math.ones(1))
      self.b = bp.math.TrainVar(bp.math.ones(1))

    def __call__(self, c):
      ab = self.a * self.b
      ab2 = ab * 2
      vv = ab2 + c
      return vv.sum(), (ab, ab2)

  test = Test()
  test_grad = bp.math.grad(test, test.vars(), argnums=0, has_aux=True)
  grads, outputs = test_grad(10.)
  print('test_grad_aux1:')
  print(grads)
  print(outputs)

  assert (grads[0] == 1.).all()  # grad of 'c'
  for g in grads[1].values():  # grad of TrainVar
    assert (g == 2.).all()
  assert (outputs[0] == 1.).all()  # 'ab'
  assert (outputs[1] == 2.).all()  # 'ab2'


def test_value_and_grad_aux1():
  class Test(bp.dnn.Module):

    def __init__(self):
      super(Test, self).__init__()

      self.a = bp.math.TrainVar(bp.math.ones(1))
      self.b = bp.math.TrainVar(bp.math.ones(1))

    def __call__(self, c):
      ab = self.a * self.b
      ab2 = ab * 2
      vv = ab2 + c
      return vv.sum(), (ab, ab2)

  test = Test()
  test_grad = bp.math.value_and_grad(test, test.vars(), argnums=0, has_aux=True)
  outputs, grads  = test_grad(10.)
  print('test_value_and_grad_aux1:')
  print(grads)
  print(outputs)

  assert (grads[0] == 1.).all()  # grad of 'c'
  for g in grads[1].values():  # grad of TrainVar
    assert (g == 2.).all()
  assert (outputs[0] == 12.).all()  # 'vv.sum()'
  assert (outputs[1][0] == 1.).all()  # 'ab'
  assert (outputs[1][1] == 2.).all()  # 'ab2'



# -*- coding: utf-8 -*-

"""
The math module for whole BrainPy ecosystem.

This module provides basic mathematical operations, including:

- numpy-like array operations
- compilations of ``jit``, ``vmap``, ``pmap`` for class objects
- automatic differentiation of ``grad``, ``value_and_grad`` for class objects
- and more ...

"""


from brainpy import errors
from brainpy.math.numpy import *


# 1. backend name
# --------------------------

BACKEND_NAME = 'numpy'


def get_backend_name():
  """Get the current backend name.

  Returns
  -------
  backend : str
      The name of the current backend name.
  """
  return BACKEND_NAME


# 2. numerical precision
# --------------------------

__dt = 0.1


def set_dt(dt):
  """Set the numerical integrator precision.

  Parameters
  ----------
  dt : float
      Numerical integration precision.
  """
  assert isinstance(dt, float), f'"dt" must a float, but we got {dt}'
  global __dt
  __dt = dt


def get_dt():
  """Get the numerical integrator precision.

  Returns
  -------
  dt : float
      Numerical integration precision.
  """
  return __dt


# 3. backend setting
# ------------------


def use_backend(name, module=None):
  # check name
  if not isinstance(name, str):
    raise errors.BrainPyError(f'"name" must be a str, but we got {type(name)}: {name}')

  # check module
  if module is None:
    if name == 'numpy':
      from brainpy.math import numpy as module
    elif name == 'jax':
      try:
        from brainpy.math import jax as module
      except ModuleNotFoundError:
        raise errors.PackageMissingError('"jax" backend need JAX, but is not installed. '
                                         'Please install jax via:\n\n'
                                         '>>> pip install jax\n'
                                         '>>> # or \n'
                                         '>>> conda install jax -c conda-forge')
    else:
      raise errors.BrainPyError(f'Unknown backend "{name}", now we only support: numpy, jax.')
  else:
    from types import ModuleType
    if not isinstance(module, ModuleType):
      raise errors.BrainPyError(f'"module" must be a module, but we got a '
                                f'type of {type(module)}: {module}')

  global_vars = globals()
  if global_vars['BACKEND_NAME'] == name:
    return

  # replace operations
  global_vars['BACKEND_NAME'] = name
  for key, value in module.__dict__.items():
    if key.startswith('_'):
      if key not in ['__name__', '__doc__', '__file__', '__path__']:
        continue
    global_vars[key] = value



# -*- coding: utf-8 -*-

from pprint import pprint

from brainpy import math, tools
from brainpy.simulation.utils import run_model

__all__ = [
  'Trajectory',
]


class Trajectory(object):
  def __init__(self, size, integrals, target_vars, fixed_vars,
               pars_update, scope, show_code=False):
    """Trajectory Class.

    Parameters
    ----------
    size : int, tuple, list
        The network size.
    integrals : list of functions, function
        The integral functions.
    target_vars : dict
        The target variables, with the format of "{key: initial_v}".
    fixed_vars : dict
        The fixed variables, with the format of "{key: fixed_v}".
    pars_update : dict
        The parameters to update.
    scope :
    """
    if callable(integrals):
      integrals = (integrals,)
    elif isinstance(integrals, (list, tuple)) and callable(integrals[0]):
      integrals = tuple(integrals)
    else:
      raise ValueError
    self.integrals = integrals
    self.target_vars = target_vars
    self.fixed_vars = fixed_vars
    self.pars_update = pars_update
    self.scope = {key: val for key, val in scope.items()}
    self.show_code = show_code

    # check network size
    if isinstance(size, int):
      size = (size,)
    elif isinstance(size, (tuple, list)):
      assert isinstance(size[0], int)
      size = tuple(size)
    else:
      raise ValueError

    # monitors, variables, parameters
    self.mon = tools.DictPlus()
    self.vars_and_pars = tools.DictPlus()
    for key, val in target_vars.items():
      self.vars_and_pars[key] = math.ones(size) * val
      self.mon[key] = []
    for key, val in fixed_vars.items():
      self.vars_and_pars[key] = math.ones(size) * val
    for key, val in pars_update.items():
      self.vars_and_pars[key] = val
    self.scope['VP'] = self.vars_and_pars
    self.scope['MON'] = self.mon
    self.scope['_fixed_vars'] = fixed_vars

    code_lines = ['def run_func(_t, _dt):']
    for integral in integrals:
      func_name = integral.__name__
      self.scope[func_name] = integral
      # update the step function
      assigns = [f'VP["{var}"]' for var in integral.brainpy_data['variables']]
      calls = [f'VP["{var}"]' for var in integral.brainpy_data['variables']]
      calls.append('_t')
      calls.extend([f'VP["{var}"]' for var in integral.brainpy_data['parameters'][1:]])
      code_lines.append(f'  {", ".join(assigns)} = {func_name}({", ".join(calls)})')
      # reassign the fixed variables
      for key, val in fixed_vars.items():
        code_lines.append(f'  VP["{key}"][:] = _fixed_vars["{key}"]')
    # monitor the target variables
    for key in target_vars.keys():
      code_lines.append(f'  MON["{key}"].append(VP["{key}"])')
    # compile
    code = '\n'.join(code_lines)
    if show_code:
      print(code)
      print()
      pprint(self.scope)
      print()

    # recompile
    exec(compile(code, '', 'exec'), self.scope)
    self.run_func = self.scope['run_func']

  def run(self, duration, report=0.1):
    if isinstance(duration, (int, float)):
      duration = [0, duration]
    elif isinstance(duration, (tuple, list)):
      assert len(duration) == 2
      duration = tuple(duration)
    else:
      raise ValueError

    # get the times
    times = math.arange(duration[0], duration[1], math.get_dt())
    # reshape the monitor
    for key in self.mon.keys():
      self.mon[key] = []
    # run the model
    run_model(run_func=self.run_func, times=times, report=report)
    # reshape the monitor
    for key in self.mon.keys():
      self.mon[key] = math.asarray(self.mon[key])




# -*- coding: utf-8 -*-

import numpy as np

from brainpy import tools

__all__ = [
  'brentq',
  'find_root_of_1d',
  'find_root_of_2d'
]

_ECONVERGED = 0
_ECONVERR = -1


@tools.numba_jit
def brentq(f, a, b, args=(), xtol=2e-14, maxiter=200, rtol=4 * np.finfo(float).eps):
  """
  Find a root of a function in a bracketing interval using Brent's method
  adapted from Scipy's brentq.

  Uses the classic Brent's method to find a zero of the function `f` on
  the sign changing interval [a , b].

  Parameters
  ----------
  f : callable
      Python function returning a number.  `f` must be continuous.
  a : number
      One end of the bracketing interval [a,b].
  b : number
      The other end of the bracketing interval [a,b].
  args : tuple, optional(default=())
      Extra arguments to be used in the function call.
  xtol : number, optional(default=2e-12)
      The computed root ``x0`` will satisfy ``np.allclose(x, x0,
      atol=xtol, rtol=rtol)``, where ``x`` is the exact root. The
      parameter must be nonnegative.
  rtol : number, optional(default=4*np.finfo(float).eps)
      The computed root ``x0`` will satisfy ``np.allclose(x, x0,
      atol=xtol, rtol=rtol)``, where ``x`` is the exact root.
  maxiter : number, optional(default=100)
      Maximum number of iterations.
  """
  if xtol <= 0:
    raise ValueError("xtol is too small (<= 0)")
  if maxiter < 1:
    raise ValueError("maxiter must be greater than 0")

  # Convert to float
  xpre = a * 1.0
  xcur = b * 1.0

  # Conditional checks for intervals in methods involving bisection
  fpre = f(xpre, *args)
  fcur = f(xcur, *args)
  funcalls = 2

  if fpre * fcur > 0:
    raise ValueError("f(a) and f(b) must have different signs")
  root = 0.0
  status = _ECONVERR

  # Root found at either end of [a,b]
  if fpre == 0:
    root = xpre
    status = _ECONVERGED
  if fcur == 0:
    root = xcur
    status = _ECONVERGED

  root, status = root, status

  # Check for sign error and early termination
  if status == _ECONVERGED:
    itr = 0
  else:
    # Perform Brent's method
    for itr in range(maxiter):
      if fpre * fcur < 0:
        xblk = xpre
        fblk = fpre
        spre = scur = xcur - xpre
      if abs(fblk) < abs(fcur):
        xpre = xcur
        xcur = xblk
        xblk = xpre

        fpre = fcur
        fcur = fblk
        fblk = fpre

      delta = (xtol + rtol * abs(xcur)) / 2
      sbis = (xblk - xcur) / 2

      # Root found
      if fcur == 0 or abs(sbis) < delta:
        status = _ECONVERGED
        root = xcur
        itr += 1
        break

      if abs(spre) > delta and abs(fcur) < abs(fpre):
        if xpre == xblk:
          # interpolate
          stry = -fcur * (xcur - xpre) / (fcur - fpre)
        else:
          # extrapolate
          dpre = (fpre - fcur) / (xpre - xcur)
          dblk = (fblk - fcur) / (xblk - xcur)
          stry = -fcur * (fblk * dblk - fpre * dpre) / \
                 (dblk * dpre * (fblk - fpre))

        if 2 * abs(stry) < min(abs(spre), 3 * abs(sbis) - delta):
          # good short step
          spre = scur
          scur = stry
        else:
          # bisect
          spre = sbis
          scur = sbis
      else:
        # bisect
        spre = sbis
        scur = sbis

      xpre = xcur
      fpre = fcur
      if abs(scur) > delta:
        xcur += scur
      else:
        xcur += (delta if sbis > 0 else -delta)
      fcur = f(xcur, *args)
      funcalls += 1

  if status == _ECONVERR:
    raise RuntimeError("Failed to converge")

  # x, funcalls, iterations = root, funcalls, itr
  return root, funcalls, itr


@tools.numba_jit
def find_root_of_1d(f, f_points, args=(), tol=1e-8):
  """Find the roots of the given function by numerical methods.

  Parameters
  ----------
  f : callable
      The function.
  f_points : onp.ndarray, list, tuple
      The value points.

  Returns
  -------
  roots : list
      The roots.
  """
  vals = f(f_points, *args)
  fs_len = len(f_points)
  fs_sign = np.sign(vals)

  roots = []
  fl_sign = fs_sign[0]
  f_i = 1
  while f_i < fs_len and fl_sign == 0.:
    roots.append(f_points[f_i - 1])
    fl_sign = fs_sign[f_i]
    f_i += 1
  while f_i < fs_len:
    fr_sign = fs_sign[f_i]
    if fr_sign == 0.:
      roots.append(f_points[f_i])
      if f_i + 1 < fs_len:
        fl_sign = fs_sign[f_i + 1]
      else:
        break
      f_i += 2
    else:
      if not np.isnan(fr_sign) and fl_sign != fr_sign:
        root, funcalls, itr = brentq(f, f_points[f_i - 1], f_points[f_i], args)
        if abs(f(root, *args)) < tol:
          roots.append(root)
      fl_sign = fr_sign
      f_i += 1

  return roots


def find_root_of_2d(f, x_bound, y_bound, args=(), shgo_args=None,
                    fl_tol=1e-6, xl_tol=1e-4, verbose=False):
  """Find the root of a two dimensional function.

  This function is aimed to find the root of :backend:`f(x) = 0`, where :backend:`x`
  is a vector with the shape of `(2,)`.

  Parameters
  ----------
  f : callable
      he objective function to be minimized.  Must be in the form
      ``f(x, *args)``, where ``x`` is the argument in the form of a 1-D array
      and ``args`` is a tuple of any additional fixed parameters needed to
      completely specify the function.
  args : tuple, optional
      Any additional fixed parameters needed to completely specify the
      objective function.
  x_bound : sequence
      Bound for the first variable. It must be a tuple/list with the format of
      ``(min, max)``, which define the lower and upper bounds for the optimizing
      argument of `f`.
  y_bound : sequence
      Bound for the second variable. It must be a tuple/list with the format of
      ``(min, max)``, which define the lower and upper bounds for the optimizing
      argument of `f`.
  shgo_args : dict
      A dictionary which contains the arguments or the parameters of `shgo` optimizer.
      It is defined in a dictionary with fields:

          - constraints
          - n
          - iters
          - callback
          - minimizer_kwargs
          - options
          - sampling_method
  fl_tol : float
      The tolerance of the function value to recognize it as a condidate of function root point.
  xl_tol : float
      The tolerance of the l2 norm distances between this point and previous points.
      If the norm distances are all bigger than `xl_tol` means this point belong to a
      new function root point.
  verbose : bool
      Whether show the shogo results.

  Returns
  -------
  res : tuple
      The roots.
  """
  from scipy.optimize import shgo

  # 1. shgo arguments
  if shgo_args is None:
    shgo_args = dict()
  if 'sampling_method' not in shgo_args:
    shgo_args['sampling_method'] = 'sobol'
  if 'n' not in shgo_args:
    shgo_args['n'] = 400

  # 2. shgo optimization
  ret = shgo(f, [x_bound, y_bound], args, **shgo_args)
  points = np.ascontiguousarray(ret.xl)
  values = np.ascontiguousarray(ret.funl)
  if verbose:
    print(ret.xl)
    print(ret.funl)

  # 3. points
  final_points = []
  for i in range(len(values)):
    if values[i] <= fl_tol:
      # first point which is less than "fl_tol"
      if len(final_points) == 0:
        final_points.append(points[i])
        continue
      # if the l2 norm distances between points[i] and
      # previous points are all bigger than "xl_tol"
      if np.alltrue(np.linalg.norm(np.array(final_points) - points[i], axis=1) > xl_tol):
        final_points.append(points[i])
    else:
      break

  # 4. x_values, y_values
  x_values, y_values = [], []
  for p in final_points:
    x_values.append(p[0])
    y_values.append(p[1])
  x_values = np.ascontiguousarray(x_values)
  y_values = np.ascontiguousarray(y_values)

  return x_values, y_values



# -*- coding: utf-8 -*-


import numpy as np

from brainpy.simulation.brainobjects import NeuGroup
from brainpy.simulation.monitors import Monitor


class TryGroup(NeuGroup):
    target_backend = 'general'

    def __init__(self, **kwargs):
        self.a = np.ones((2, 2))
        super(TryGroup, self).__init__(size=1, **kwargs)

    def update(self, _t):
        self.a += 1


def test_non_array():
    set(dt=0.1)
    try1 = TryGroup(monitors=['a'])
    try1.a = 1.
    try1.run(100.)

    assert np.ndim(try1.mon.a) == 2 and np.shape(try1.mon.a)[1] == 1
    assert np.allclose(np.arange(2, 1002).reshape((-1, 1)), try1.mon.a)


def test_1d_array():
    set(dt=0.1)
    try1 = TryGroup(monitors=['a'])
    try1.a = np.ones(1)
    try1.run(100.)

    assert np.ndim(try1.mon.a) == 2 and np.shape(try1.mon.a)[1] == 1
    assert np.allclose(np.arange(2, 1002).reshape((-1, 1)), try1.mon.a)


def test_2d_array():
    set(dt=0.1)
    try1 = TryGroup(monitors=['a'])
    try1.a = np.ones((2, 2))
    try1.run(100.)

    assert np.ndim(try1.mon.a) == 2 and np.shape(try1.mon.a)[1] == 4
    series = np.arange(2, 1002).reshape((-1, 1))
    series = np.repeat(series, 4, axis=1)
    assert np.allclose(series, try1.mon.a)


def test_monitor_with_every():
    set(dt=0.1)

    # try1: 2d array
    try1 = TryGroup(monitors=Monitor(variables=['a'], every=[1.]))
    try1.run(100.)
    assert np.ndim(try1.mon.a) == 2 and np.shape(try1.mon.a)[1] == 4
    series = np.arange(2, 1002, 1. / 0.1).reshape((-1, 1))
    series = np.repeat(series, 4, axis=1)
    assert np.allclose(series, try1.mon.a)

    # try2: 1d array
    try2 = TryGroup(monitors=Monitor(variables=['a'], every=[1.]))
    try2.a = np.array([1., 1.])
    try2.run(100.)
    assert np.ndim(try2.mon.a) == 2 and np.shape(try2.mon.a)[1] == 2
    series = np.arange(2, 1002, 1. / 0.1).reshape((-1, 1))
    series = np.repeat(series, 2, axis=1)
    assert np.allclose(series, try2.mon.a)

    # try2: scalar
    try3 = TryGroup(monitors=Monitor(variables=['a'], every=[1.]))
    try3.a = 1.
    try3.run(100.)
    assert np.ndim(try3.mon.a) == 2 and np.shape(try3.mon.a)[1] == 1
    series = np.arange(2, 1002, 1. / 0.1).reshape((-1, 1))
    assert np.allclose(series, try3.mon.a)


# test_non_array()
# test_1d_array()
# test_2d_array()
# test_monitor_with_every()



# -*- coding: utf-8 -*-

from brainpy import errors

__all__ = [
  'check_name',
  'get_name',
]

_name2id = dict()
_typed_names = {}


def check_name(name, obj):
  if not name.isidentifier():
    raise errors.BrainPyError(f'"{name}" isn\'t a valid identifier '
                              f'according to Python language definition. '
                              f'Please choose another name.')
  if name in _name2id:
    if _name2id[name] != id(obj):
      raise errors.UniqueNameError(f'In BrainPy, each object should have a unique name. '
                                   f'However, we detect that {obj} has a used name "{name}".')
  else:
    _name2id[name] = id(obj)


def get_name(type):
  if type not in _typed_names:
    _typed_names[type] = 0
  name = f'{type}{_typed_names[type]}'
  _typed_names[type] += 1
  return name



# -*- coding: utf-8 -*-

import brainpy as bp


def test_zero_init():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    init = bp.initialize.ZeroInit()
    for size in [(100,), (10, 20), (10, 20, 30)]:
      weights = init(size)
      assert weights.shape == size
      assert isinstance(weights, bp.math.ndarray)


def test_one_init():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    for size in [(100,), (10, 20), (10, 20, 30)]:
      for value in [0., 1., -1.]:
        init = bp.initialize.OneInit(value=value)
        weights = init(size)
        assert weights.shape == size
        assert isinstance(weights, bp.math.ndarray)
        assert (weights == value).all()


def test_identity_init():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    for size in [(100,), (10, 20), ]:
      for value in [0., 1., -1.]:
        init = bp.initialize.Identity(value=value)
        weights = init(size)
        if len(size) == 1:
          assert weights.shape == (size[0], size[0])
        else:
          assert weights.shape == size
        assert isinstance(weights, bp.math.ndarray)



# -*- coding: utf-8 -*-

import brainpy as bp
from brainpy import connect


def test_one2one():
  for size in [100, (3, 4), (4, 5, 6)]:
    conn = connect.One2One()(pre_size=size, post_size=size)
    mat = conn.require(connect.CONN_MAT)
    res = conn.require('pre_ids', 'post_ids', 'pre2post', 'pre_slice')
    num = bp.simulation.size2len(size)

    actual_mat = bp.math.zeros((num, num), dtype=bp.math.bool_)
    actual_mat = bp.math.fill_diagonal(actual_mat, True)
    assert bp.math.array_equal(actual_mat, mat)
    assert bp.math.array_equal(res[0], bp.math.arange(num))
    assert bp.math.array_equal(res[1], bp.math.arange(num))

    print()
    print('pre_ids', res[0])
    print('post_ids', res[1])
    print('pre2post', res[2])
    print('pre_slice', res[3])


def test_all2all():
  for has_self in [True, False]:
    for size in [100, (3, 4), (4, 5, 6)]:
      conn = connect.All2All(include_self=has_self)(pre_size=size, post_size=size)
      mat = conn.require(connect.CONN_MAT)
      res = conn.require('pre_ids', 'post_ids', 'pre2post', 'pre_slice')
      num = bp.simulation.size2len(size)

      actual_mat = bp.math.ones((num, num), dtype=bp.math.bool_)
      if not has_self:
        actual_mat = bp.math.fill_diagonal(actual_mat, False)
      assert bp.math.array_equal(actual_mat, mat)

      print()
      print('pre_ids', res[0])
      print('post_ids', res[1])
      print('pre2post', res[2])
      print('pre_slice', res[3])


def test_grid_four():
  pass




# -*- coding: utf-8 -*-

from brainpy import math, errors
from brainpy.base.base import Base
from brainpy.base.collector import Collector
from brainpy.simulation import utils
from brainpy.simulation.monitor import Monitor

ConstantDelay = None

__all__ = [
  'DynamicalSystem',
  'Container',
]

_error_msg = 'Unknown model type: {type}. ' \
             'Currently, BrainPy only supports: \n' \
             '1. function \n' \
             '2. function name (str) \n' \
             '3. tuple/dict of functions \n' \
             '4. tuple of function names \n'


class DynamicalSystem(Base):
  """Base Dynamical System class.

  Any object has step functions will be a dynamical system.
  That is to say, in BrainPy, the essence of the dynamical system
  is the "step functions".

  Parameters
  ----------
  steps : tuple of str, tuple of function, dict of (str, function), optional
      The callable function, or a list of callable functions.
  monitors : None, list, tuple, datastructures.Monitor
      Variables to monitor.
  name : str, optional
      The name of the dynamic system.
  """

  def __init__(self, steps=None, monitors=None, name=None):
    super(DynamicalSystem, self).__init__(name=name)

    # step functions
    if steps is None:
      steps = ('update',)
    self.steps = Collector()
    if isinstance(steps, tuple):
      for step in steps:
        if isinstance(step, str):
          self.steps[step] = getattr(self, step)
        elif callable(step):
          self.steps[step.__name__] = step
        else:
          raise errors.BrainPyError(_error_msg.format(type(steps[0])))
    elif isinstance(steps, dict):
      for key, step in steps.items():
        if callable(step):
          self.steps[key] = step
        else:
          raise errors.BrainPyError(_error_msg.format(type(step)))
    else:
      raise errors.BrainPyError(_error_msg.format(type(steps)))

    # monitors
    if monitors is None:
      self.mon = Monitor(target=self, variables=[])
    elif isinstance(monitors, (list, tuple, dict)):
      self.mon = Monitor(target=self, variables=monitors)
    elif isinstance(monitors, Monitor):
      self.mon = monitors
      self.mon.target = self
    else:
      raise errors.BrainPyError(f'"monitors" only supports list/tuple/dict/ '
                                f'instance of Monitor, not {type(monitors)}.')

    # runner and run function
    self._input_step = None
    self._monitor_step = None

    # batch size
    self.num_batch = None

  def register_constant_delay(self, key, size, delay, dtype=None):
    """Register a constant delay.

    Parameters
    ----------
    key : str
      The delay name.
    size : int, list of int, tuple of int
      The delay data size.
    delay : int, float, ndarray
      The delay time, with the unit same with `brainpy.math.get_dt()`.
    dtype : optional
      The data type.

    Returns
    -------
    delay : ConstantDelay
        An instance of ConstantDelay.
    """
    global ConstantDelay
    if ConstantDelay is None: from brainpy.simulation.brainobjects.delays import ConstantDelay

    if not hasattr(self, 'steps'):
      raise errors.BrainPyError('Please initialize the super class first before '
                                'registering constant_delay. \n\n'
                                'super(YourClassName, self).__init__(**kwargs)')
    if not key.isidentifier(): raise ValueError(f'{key} is not a valid identifier.')
    cdelay = ConstantDelay(size=size,
                           delay=delay,
                           num_batch=getattr(self, 'num_batch', None),
                           name=f'{self.name}_delay_{key}',
                           dtype=dtype)
    self.steps[f'{key}_update'] = cdelay.update
    return cdelay

  def update(self, _t, _dt, **kwargs):
    """The function to specify the updating rule.

    Parameters
    ----------
    _t : float
      The current time.
    _dt : float
      The time step.
    """
    raise NotImplementedError('Must implement "update" function by user self.')

  def run(self, duration, dt=None, report=0., inputs=(), extra_func=None):
    """The running function.

    Parameters
    ----------
    inputs : list, tuple
      The inputs for this instance of DynamicalSystem. It should the format
      of `[(target, value, [type, operation])]`, where `target` is the
      input target, `value` is the input value, `type` is the input type
      (such as "fix" or "iter"), `operation` is the operation for inputs
      (such as "+", "-", "*", "/", "=").

      - ``target``: should be a string. Can be specified by the *absolute access* or *relative access*.
      - ``value``: should be a scalar, vector, matrix, iterable function or objects.
      - ``type``: should be a string. "fix" means the input `value` is a constant. "iter" means the
        input `value` can be changed over time.
      - ``operation``: should be a string, support `+`, `-`, `*`, `/`, `=`.
      - Also, if you want to specify multiple inputs, just give multiple ``(target, value, [type, operation])``,
        for example ``[(target1, value1), (target2, value2)]``.

    duration : float, int, tuple, list
      The running duration.

    report : float
      The percent of progress to report. [0, 1]. If zero, the model
      will not output report progress.

    dt : float, optional
      The numerical integration step size.

    extra_func : function, callable
      The extra function to run during each time step.

    Returns
    -------
    running_time : float
      The total running time.
    """

    if dt is None:
      dt = math.get_dt()
    assert isinstance(dt, (int, float))

    # 2. Build the inputs.
    #    All the inputs are wrapped into a single function.
    self._input_step = utils.build_input_func(
      utils.check_and_format_inputs(host=self, inputs=inputs), show_code=False)

    # 3. Build the monitors.
    #    All the monitors are wrapped in a single function.
    if self._monitor_step is None:
      self._monitor_step = utils.build_monitor_func(
        utils.check_and_format_monitors(host=self), show_code=False)
    else:
      for node in self.nodes().unique().values():
        for key in node.mon.item_contents.keys():
          node.mon.item_contents[key] = []  # reshape the monitor items

    # 4. times
    start, end = utils.check_duration(duration)
    times = math.arange(start, end, dt)

    # 5. run the model
    # ----
    # 5.1 iteratively run the step function.
    # 5.2 report the running progress.
    # 5.3 return the overall running time.

    def step_run(_t, _dt, **kwargs):
      self._input_step(_t=_t, _dt=_dt)
      for step in self.steps.values():
        step(_t=_t, _dt=_dt, **kwargs)
      self._monitor_step(_t=_t, _dt=_dt)

    running_time = utils.run_model(run_func=step_run,
                                   times=times,
                                   report=report,
                                   dt=dt,
                                   extra_func=extra_func)

    # 6. monitor
    # --
    # 6.1 add 'ts' variable to every monitor
    # 6.2 wrap the monitor iterm with the 'list' type into the 'ndarray' type
    for node in self.nodes().unique().values():
      if node.mon.num_item > 0:
        node.mon.ts = times
        for key, val in list(node.mon.item_contents.items()):
          val = math.asarray(node.mon.item_contents[key])
          node.mon.item_contents[key] = val

    return running_time

  def find_fixed_points(self):
    pass

  def evaluate_stability(self):
    pass

  def continuation(self, method='euler_newton'):
    pass


class Container(DynamicalSystem):
  """Container object which is designed to add other instances of DynamicalSystem.

  Parameters
  ----------
  steps : tuple of function, tuple of str, dict of (str, function), optional
      The step functions.
  monitors : tuple, list, Monitor, optional
      The monitor object.
  name : str, optional
      The object name.
  show_code : bool
      Whether show the formatted code.
  ds_dict : dict of (str, )
      The instance of DynamicalSystem with the format of "key=dynamic_system".
  """

  def __init__(self, *ds_tuple, steps=None, monitors=None, name=None, **ds_dict):
    # children dynamical systems
    self.child_ds = dict()
    for ds in ds_tuple:
      if not isinstance(ds, DynamicalSystem):
        raise errors.BrainPyError(f'{self.__class__.__name__} receives instances of '
                                  f'DynamicalSystem, however, we got {type(ds)}.')
      if ds.name in self.child_ds:
        raise ValueError(f'{ds.name} has been paired with {ds}. Please change a unique name.')
      self.child_ds[ds.name] = ds
    for key, ds in ds_dict.items():
      if not isinstance(ds, DynamicalSystem):
        raise errors.BrainPyError(f'{self.__class__.__name__} receives instances of '
                                  f'DynamicalSystem, however, we got {type(ds)}.')
      if key in self.child_ds:
        raise ValueError(f'{key} has been paired with {ds}. Please change a unique name.')
      self.child_ds[key] = ds
    # step functions in children dynamical systems
    self.child_steps = dict()
    for ds_key, ds in self.child_ds.items():
      for step_key, step in ds.steps.items():
        self.child_steps[f'{ds_key}_{step_key}'] = step

    # integrative step function
    if steps is None:
      steps = ('update',)
    super(Container, self).__init__(steps=steps, monitors=monitors, name=name)

  def update(self, _t, _dt):
    """Step function of a network.

    In this update function, the step functions in children systems are
    iteratively called.
    """
    for step in self.child_steps.values():
      step(_t, _dt)

  def nodes(self, method='absolute', _paths=None):
    if _paths is None:
      _paths = set()
    gather = self._nodes_in_container(self.child_ds, method=method, _paths=_paths)
    gather.update(super(Container, self).nodes(method=method, _paths=_paths))
    return gather

  def __getattr__(self, item):
    children_ds = super(Container, self).__getattribute__('child_ds')
    if item in children_ds:
      return children_ds[item]
    else:
      return super(Container, self).__getattribute__(item)



# -*- coding: utf-8 -*-

import logging

import numpy as np

from brainpy import math, errors

try:
  import numba as nb
except ModuleNotFoundError:
  nb = None

logger = logging.getLogger('brainpy.simulation.connect')

__all__ = [
  'ij2mat',
  'mat2ij',
  'pre2post',
  'post2pre',
  'pre2syn',
  'post2syn',
  'pre_slice',
  'post_slice',
]


def _numba_backend():
  r = math.get_backend_name().startswith('numpy')
  return r and (nb is not None)


def ij2mat(i, j, num_pre=None, num_post=None):
  """Convert i-j connection to matrix connection.

  Parameters
  ----------
  i : list, np.ndarray
      Pre-synaptic neuron index.
  j : list, np.ndarray
      Post-synaptic neuron index.
  num_pre : int, optional
      The number of the pre-synaptic neurons.
  num_post : int, optional
      The number of the post-synaptic neurons.

  Returns
  -------
  conn_mat : np.ndarray
      A 2D ndarray connectivity matrix.
  """
  if len(i) != len(j):
    raise errors.BrainPyError('"i" and "j" must be the equal length.')
  if num_pre is None:
    logger.warning('"num_pre" is not provided, the result may not be accurate.')
    num_pre = math.max(i)
  if num_post is None:
    logger.warning('"num_post" is not provided, the result may not be accurate.')
    num_post = math.max(j)
  conn_mat = math.zeros((num_pre, num_post), dtype=math.bool_)
  conn_mat[i, j] = True
  return conn_mat


def mat2ij(conn_mat):
  """Get the i-j connections from connectivity matrix.

  Parameters
  ----------
  conn_mat : np.ndarray
      Connectivity matrix with `(num_pre, num_post)` shape.

  Returns
  -------
  conn_tuple : tuple
      (Pre-synaptic neuron indexes,
       post-synaptic neuron indexes).
  """
  if len(math.shape(conn_mat)) != 2:
    raise errors.BrainPyError('Connectivity matrix must be in the '
                              'shape of (num_pre, num_post).')
  pre_ids, post_ids = math.where(conn_mat > 0)
  return math.array(pre_ids, dtype=math.int_), math.array(post_ids, dtype=math.int_)


def pre2post(i, j, num_pre=None):
  """Get pre2post connections from `i` and `j` indexes.

  Parameters
  ----------
  i : list, np.ndarray
      The pre-synaptic neuron indexes.
  j : list, np.ndarray
      The post-synaptic neuron indexes.
  num_pre : int, None
      The number of the pre-synaptic neurons.

  Returns
  -------
  conn : list
      The conn list of pre2post.
  """
  if len(i) != len(j):
    raise errors.BrainPyError('The length of "i" and "j" must be the same.')
  if num_pre is None:
    logger.warning('"num_pre" is not provided, the result may not be accurate.')
    num_pre = math.max(i)

  pre2post_list = [[] for _ in range(num_pre)]
  for pre_id, post_id in zip(i, j):
    pre2post_list[pre_id].append(post_id)
  pre2post_list = [math.array(l, dtype=math.int_) for l in pre2post_list]

  if _numba_backend():
    pre2post_list_nb = nb.typed.List()
    for pre_id in range(num_pre):
      pre2post_list_nb.append(pre2post_list[pre_id])
    pre2post_list = math.Variable(math.asarray(pre2post_list, dtype=object),
                                  type='connection')
    pre2post_list.value = pre2post_list_nb
  return pre2post_list


def post2pre(i, j, num_post=None):
  """Get post2pre connections from `i` and `j` indexes.

  Parameters
  ----------
  i : list, np.ndarray
      The pre-synaptic neuron indexes.
  j : list, np.ndarray
      The post-synaptic neuron indexes.
  num_post : int, None
      The number of the post-synaptic neurons.

  Returns
  -------
  conn : list
      The conn list of post2pre.
  """

  if len(i) != len(j):
    raise errors.BrainPyError('The length of "i" and "j" must be the same.')
  if num_post is None:
    logger.warning('WARNING: "num_post" is not provided, the result may not be accurate.')
    num_post = math.max(j)

  post2pre_list = [[] for _ in range(num_post)]
  for pre_id, post_id in zip(i, j):
    post2pre_list[post_id].append(pre_id)
  post2pre_list = [math.array(l, dtype=math.int_) for l in post2pre_list]

  if _numba_backend():
    post2pre_list_nb = nb.typed.List()
    for post_id in range(num_post):
      post2pre_list_nb.append(post2pre_list[post_id])
    post2pre_list = math.Variable(math.asarray(post2pre_list, dtype=object),
                                  type='connection')
    post2pre_list.value = post2pre_list_nb
  return post2pre_list


def pre2syn(i, num_pre=None):
  """Get pre2syn connections from `i` and `j` indexes.

  Parameters
  ----------
  i : list, np.ndarray
      The pre-synaptic neuron indexes.
  num_pre : int
      The number of the pre-synaptic neurons.

  Returns
  -------
  conn : list
      The conn list of pre2syn.
  """
  if num_pre is None:
    logger.warning('WARNING: "num_pre" is not provided, the result may not be accurate.')
    num_pre = math.max(i)

  pre2syn_list = [[] for _ in range(num_pre)]
  for syn_id, pre_id in enumerate(i):
    pre2syn_list[pre_id].append(syn_id)
  pre2syn_list = [math.array(l, dtype=math.int_) for l in pre2syn_list]

  if _numba_backend():
    pre2syn_list_nb = nb.typed.List()
    for pre_ids in pre2syn_list:
      pre2syn_list_nb.append(pre_ids)
    pre2syn_list = math.Variable(math.asarray(pre2syn_list, dtype=object),
                                 type='connection')
    pre2syn_list.value = pre2syn_list_nb
  return pre2syn_list


def post2syn(j, num_post=None):
  """Get post2syn connections from `i` and `j` indexes.

  Parameters
  ----------
  j : list, np.ndarray
      The post-synaptic neuron indexes.
  num_post : int
      The number of the post-synaptic neurons.

  Returns
  -------
  conn : list
      The conn list of post2syn.
  """
  if num_post is None:
    logger.warning('WARNING: "num_post" is not provided, the result may not be accurate.')
    num_post = math.max(j)

  post2syn_list = [[] for _ in range(num_post)]
  for syn_id, post_id in enumerate(j):
    post2syn_list[post_id].append(syn_id)
  post2syn_list = [math.array(l, dtype=math.int_) for l in post2syn_list]

  if _numba_backend():
    post2syn_list_nb = nb.typed.List()
    for pre_ids in post2syn_list:
      post2syn_list_nb.append(pre_ids)
    post2syn_list = math.Variable(math.asarray(post2syn_list, dtype=object),
                                  type='connection')
    post2syn_list.value = post2syn_list_nb

  return post2syn_list


def pre_slice(i, j, num_pre=None):
  """Get post slicing connections by pre-synaptic ids.

  Parameters
  ----------
  i : list, np.ndarray
      The pre-synaptic neuron indexes.
  j : list, np.ndarray
      The post-synaptic neuron indexes.
  num_pre : int
      The number of the pre-synaptic neurons.

  Returns
  -------
  conn : list
      The conn list of post2syn.
  """
  # check
  if len(i) != len(j):
    raise errors.BrainPyError('The length of "i" and "j" must be the same.')
  if num_pre is None:
    logger.warning('WARNING: "num_pre" is not provided, the result may not be accurate.')
    num_pre = math.max(i)

  # pre2post connection
  pre2post_list = [[] for _ in range(num_pre)]
  for pre_id, post_id in zip(i, j):
    pre2post_list[pre_id].append(post_id)
  pre_ids, post_ids = [], []
  for pre_i, posts in enumerate(pre2post_list):
    post_ids.extend(posts)
    pre_ids.extend([pre_i] * len(posts))
  post_ids = math.array(post_ids, dtype=math.int_)
  pre_ids = math.array(pre_ids, dtype=math.int_)

  # pre2post slicing
  slicing = []
  start = 0
  for posts in pre2post_list:
    end = start + len(posts)
    slicing.append([start, end])
    start = end
  slicing = math.array(slicing, dtype=math.int_)

  return pre_ids, post_ids, slicing


def post_slice(i, j, num_post=None):
  """Get pre slicing connections by post-synaptic ids.

  Parameters
  ----------
  i : list, np.ndarray
      The pre-synaptic neuron indexes.
  j : list, np.ndarray
      The post-synaptic neuron indexes.
  num_post : int
      The number of the post-synaptic neurons.

  Returns
  -------
  conn : list
      The conn list of post2syn.
  """
  if len(i) != len(j):
    raise errors.BrainPyError('The length of "i" and "j" must be the same.')
  if num_post is None:
    logger.warning('WARNING: "num_post" is not provided, the result may not be accurate.')
    num_post = math.max(j)

  # post2pre connection
  post2pre_list = [[] for _ in range(num_post)]
  for pre_id, post_id in zip(i, j):
    post2pre_list[post_id].append(pre_id)
  pre_ids, post_ids = [], []
  for _post_id, _pre_ids in enumerate(post2pre_list):
    pre_ids.extend(_pre_ids)
    post_ids.extend([_post_id] * len(_pre_ids))
  post_ids = math.array(post_ids, dtype=math.int_)
  pre_ids = math.array(pre_ids, dtype=math.int_)

  # post2pre slicing
  slicing = []
  start = 0
  for pres in post2pre_list:
    end = start + len(pres)
    slicing.append([start, end])
    start = end
  slicing = math.array(slicing, dtype=math.int_)

  return pre_ids, post_ids, slicing
